from __future__ import annotations
import asyncio
from dataclasses import dataclass
import inspect
import logging
from typing import Callable, Dict, List, Optional, Tuple, Type, TYPE_CHECKING, Union

from .room.model import Room, RoomMessage
from .user.model import ChatMessage, User
from .protocol.primitives import (
    DirectoryData,
    MessageDataclass,
    Recommendation,
)
from .protocol.messages import (
    AddUser,
    AdminMessage,
    CheckPrivileges,
    GetGlobalRecommendations,
    GetItemRecommendations,
    GetItemSimilarUsers,
    GetRecommendations,
    GetSimilarUsers,
    GetUserInterests,
    GetUserStats,
    GetUserStatus,
    JoinRoom,
    Kicked,
    LeaveRoom,
    Login,
    PeerDirectoryContentsReply,
    PeerSharesReply,
    PeerUserInfoReply,
    PrivateChatMessage,
    PrivateRoomGrantMembership,
    PrivateRoomGrantOperator,
    PrivateRoomMembers,
    PrivateRoomMembershipGranted,
    PrivateRoomMembershipRevoked,
    PrivateRoomOperatorGranted,
    PrivateRoomOperatorRevoked,
    PrivateRoomOperators,
    PrivateRoomRevokeMembership,
    PrivateRoomRevokeOperator,
    PrivilegedUsers,
    PublicChatMessage,
    RoomChatMessage,
    RoomList,
    RoomTickerAdded,
    RoomTickerRemoved,
    RoomTickers,
    UserJoinedRoom,
    UserLeftRoom,
)
from .search.model import SearchRequest, SearchResult
from .session import Session

if TYPE_CHECKING:
    from .network.connection import (
        Connection,
        ConnectionState,
        CloseReason,
        PeerConnection,
    )
    from .transfer.model import Transfer, TransferProgressSnapshot


logger = logging.getLogger(__name__)


# Internal functions

def on_message(message_class: Type[MessageDataclass]):
    """Decorator for methods listening to specific `MessageData` events"""
    def register(event_func):
        event_func._registered_message = message_class
        return event_func

    return register


def build_message_map(obj: object) -> Dict[Type[MessageDataclass], Callable]:
    methods = inspect.getmembers(obj, predicate=inspect.ismethod)
    mapping: Dict[Type[MessageDataclass], Callable] = {}
    for _, method in methods:
        registered_message = getattr(method, '_registered_message', None)
        if registered_message:
            mapping[registered_message] = method
    return mapping


# External functions

class EventBus:

    def __init__(self):
        self._events: Dict[Type[Event], List[Tuple[int, Callable[[Event], None]]]] = {}

    def register(self, event_class: Type[Event], listener: Callable[[Event], None], priority: int = 100):
        entry = (priority, listener)
        try:
            self._events[event_class].append(entry)
        except KeyError:
            self._events[event_class] = [entry, ]

        self._events[event_class].sort(key=lambda e: e[0])

    async def emit(self, event: Event):
        try:
            listeners = self._events[event.__class__]
        except KeyError:
            pass
        else:
            for _, listener in listeners:
                try:
                    if asyncio.iscoroutinefunction(listener):
                        await listener(event)
                    else:
                        listener(event)
                except Exception:
                    logger.exception(f"exception notifying listener {listener!r} of event {event!r}")


# Public events
class Event:
    """Base class for events"""


@dataclass(frozen=True)
class SessionInitializedEvent(Event):
    """Emitted after successful login"""
    session: Session
    raw_message: Login.Response


@dataclass(frozen=True)
class SessionDestroyedEvent(Event):
    """Emitted after the login session has been destroyed (this is always after
    disconnect)
    """
    session: Session


@dataclass(frozen=True)
class AdminMessageEvent(Event):
    """Emitted when a global admin message has been received"""
    message: str
    raw_message: AdminMessage.Response


@dataclass(frozen=True)
class KickedEvent(Event):
    """Emitted when we are kicked from the server"""
    raw_message: Kicked.Response


@dataclass(frozen=True)
class UserTrackingEvent(Event):
    """Emitted when a user is now successfully by the library tracked. This will
    be triggered when:

    The `user` object will contain the tracked user. `data` will contain the
    data returned by the server
    """
    user: User
    raw_message: AddUser.Response


@dataclass(frozen=True)
class UserTrackingFailedEvent(Event):
    username: str
    raw_message: AddUser.Response


@dataclass(frozen=True)
class UserUntrackingEvent(Event):
    """Emitted when a user is no longer tracked by the library"""
    user: User


@dataclass(frozen=True)
class UserStatusUpdateEvent(Event):
    """Emitted when the server sends an update for the users status / privileges
    or a request for it

    Following
    """
    before: User
    current: User
    raw_message: GetUserStatus.Response


@dataclass(frozen=True)
class UserStatsUpdateEvent(Event):
    """Emitted when the server send an update for the user's sharing statistics
    """
    before: User
    current: User
    raw_message: GetUserStats.Response


@dataclass(frozen=True)
class UserInfoUpdateEvent(Event):
    """Emitted when user info has been received"""
    before: User
    current: User
    raw_message: PeerUserInfoReply.Request


@dataclass(frozen=True)
class RoomListEvent(Event):
    rooms: List[Room]
    raw_message: RoomList.Response


@dataclass(frozen=True)
class RoomMessageEvent(Event):
    message: RoomMessage
    raw_message: RoomChatMessage.Response


@dataclass(frozen=True)
class RoomTickersEvent(Event):
    """Emitted when a list of tickers has been received for a room"""
    room: Room
    tickers: Dict[str, str]
    raw_message: RoomTickers.Response


@dataclass(frozen=True)
class RoomTickerAddedEvent(Event):
    """Emitted when a ticker has been added to the room by a user"""
    room: Room
    user: User
    ticker: str
    raw_message: RoomTickerAdded.Response


@dataclass(frozen=True)
class RoomTickerRemovedEvent(Event):
    """Emitted when a ticker has been removed from the room by a user"""
    room: Room
    user: User
    raw_message: RoomTickerRemoved.Response


@dataclass(frozen=True)
class RoomJoinedEvent(Event):
    """Emitted after a user joined a chat room

    The value of `user` will be `None` in case it is us who has joined the room
    """
    room: Room
    raw_message: Union[JoinRoom.Response, UserJoinedRoom.Response]
    user: Optional[User] = None


@dataclass(frozen=True)
class RoomLeftEvent(Event):
    """Emitted after a user left a chat room

    The value of `user` will be `None` in case it is us who has left the room
    """
    room: Room
    raw_message: Union[LeaveRoom.Response, UserLeftRoom.Response]
    user: Optional[User] = None


@dataclass(frozen=True)
class RoomMembershipGrantedEvent(Event):
    """Emitted when a member has been added to the private room

    The value of `member` will be `None` in case it is us who has been added to
    the room
    """
    room: Room
    raw_message: Union[PrivateRoomGrantMembership.Response, PrivateRoomMembershipGranted.Response]
    member: Optional[User] = None


@dataclass(frozen=True)
class RoomMembershipRevokedEvent(Event):
    """Emitted when a member has been removed to the private room

    The value of `member` will be `None` in case it is us who has been removed
    from the room
    """
    room: Room
    raw_message: Union[PrivateRoomRevokeMembership.Response, PrivateRoomMembershipRevoked.Response]
    member: Optional[User] = None


@dataclass(frozen=True)
class RoomOperatorGrantedEvent(Event):
    """Emitted when a member has been granted operator privileges on a private
    room

    The value of `member` will be `None` in case it is us who has been granted
    operator
    """
    room: Room
    raw_message: Union[PrivateRoomGrantOperator.Response, PrivateRoomOperatorGranted.Response]
    member: Optional[User] = None


@dataclass(frozen=True)
class RoomOperatorRevokedEvent(Event):
    """Emitted when a member had operator privileges revoked on a private room

    The value of `user` will be `None` in case it is us who has been revoked
    operator
    """
    room: Room
    raw_message: Union[PrivateRoomRevokeOperator.Response, PrivateRoomOperatorRevoked.Response]
    member: Optional[User] = None


@dataclass(frozen=True)
class RoomOperatorsEvent(Event):
    """Emitted when the server sends us a list of operators in a room"""
    room: Room
    operators: List[User]
    raw_message: PrivateRoomOperators.Response


@dataclass(frozen=True)
class RoomMembersEvent(Event):
    """Emitted when the server sends us a list of members of a room. The list
    of members always excludes the owner of the room
    """
    room: Room
    members: List[User]
    raw_message: PrivateRoomMembers.Response


@dataclass(frozen=True)
class PrivateMessageEvent(Event):
    """Emitted when a private message has been received"""
    message: ChatMessage
    raw_message: PrivateChatMessage.Response


@dataclass(frozen=True)
class PublicMessageEvent(Event):
    """Emitted when a message has been received from a public room and public
    chat has been enabled
    """
    timestamp: int
    user: User
    room: Room
    message: str
    raw_message: PublicChatMessage.Response


@dataclass(frozen=True)
class SearchResultEvent(Event):
    """Emitted when a search result has been received"""
    query: SearchRequest
    result: SearchResult


@dataclass(frozen=True)
class SearchRequestReceivedEvent(Event):
    """Emitted when a search request by another user has been received"""
    username: str
    query: str
    result_count: int


@dataclass(frozen=True)
class SimilarUsersEvent(Event):
    users: List[User]
    raw_message: Union[GetSimilarUsers.Response, GetItemSimilarUsers.Response]
    item: Optional[str] = None


@dataclass(frozen=True)
class RecommendationsEvent(Event):
    recommendations: List[Recommendation]
    unrecommendations: List[Recommendation]
    raw_message: GetRecommendations.Response


@dataclass(frozen=True)
class GlobalRecommendationsEvent(Event):
    recommendations: List[Recommendation]
    unrecommendations: List[Recommendation]
    raw_message: GetGlobalRecommendations.Response


@dataclass(frozen=True)
class ItemRecommendationsEvent(Event):
    item: str
    recommendations: List[Recommendation]
    raw_message: GetItemRecommendations.Response


@dataclass(frozen=True)
class UserInterestsEvent(Event):
    user: User
    interests: List[str]
    hated_interests: List[str]
    raw_message: GetUserInterests.Response


@dataclass(frozen=True)
class PrivilegedUsersEvent(Event):
    """Emitted when the list of privileged users has been received"""
    users: List[User]
    raw_message: PrivilegedUsers.Response


@dataclass(frozen=True)
class PrivilegedUserAddedEvent(Event):
    """Emitted when a new privileged user has been added"""
    user: User


@dataclass(frozen=True)
class PrivilegesUpdateEvent(Event):
    """Emitted when we receive a message containing how much time is left on the
    current user's privileges. If `time_left` is 0 the current user has no
    privileges
    """
    time_left: int
    raw_message: CheckPrivileges.Response


# Peer

@dataclass(frozen=True)
class UserSharesReplyEvent(Event):
    user: User
    directories: List[DirectoryData]
    locked_directories: List[DirectoryData]
    raw_message: PeerSharesReply.Request


@dataclass(frozen=True)
class UserDirectoryEvent(Event):
    user: User
    directory: str
    directories: List[DirectoryData]
    raw_message: PeerDirectoryContentsReply.Request


# Transfer

@dataclass(frozen=True)
class TransferAddedEvent(Event):
    """Emitted when a transfer has been attached to the client"""
    transfer: Transfer


@dataclass(frozen=True)
class TransferRemovedEvent(Event):
    """Emitted when a transfer has been detached from the client"""
    transfer: Transfer


@dataclass(frozen=True)
class TransferProgressEvent(Event):
    """Called periodically to report progress. The interval is determined by the
    `transfers.progress_interval` settings parameter

    This event only includes transfers where the transfer state has been changed
    since the previous event. If there are no updates the event will no be
    called
    """
    updates: List[Tuple[Transfer, TransferProgressSnapshot, TransferProgressSnapshot]]
    """List of progress updates: transfer instance, previous snapshot, current
    snapshot
    """


# Internal Events

class InternalEvent(Event):
    """Base class for internal events"""


@dataclass(frozen=True)
class ConnectionStateChangedEvent(InternalEvent):
    """Indicates a state change in any type of connection"""
    connection: Connection
    state: ConnectionState
    close_reason: Optional[CloseReason] = None


@dataclass(frozen=True)
class ServerReconnectedEvent(InternalEvent):
    """Indicates server reconnected, used internally to automatically log back
    in
    """


@dataclass(frozen=True)
class MessageReceivedEvent(InternalEvent):
    """Emitted when a message was received on a connection"""
    message: MessageDataclass
    connection: Connection


@dataclass(frozen=True)
class PeerInitializedEvent(InternalEvent):
    """Emitted when a new peer connection has been established and the
    initialization message has been received or sent
    """
    connection: PeerConnection
    requested: bool
    """Indictes whether the connection was initialized by another user or opened
    on the client's request
    """


@dataclass(frozen=True)
class ScanCompleteEvent(InternalEvent):
    """Emitted when shares scan was completed"""
    folder_count: int
    file_count: int



from __future__ import annotations
import asyncio
from enum import auto, Enum
from typing import BinaryIO, Callable, List, Optional, TYPE_CHECKING, Union
import logging
import socket
import struct

from ..constants import (
    PEER_CONNECT_TIMEOUT,
    PEER_READ_TIMEOUT,
    SERVER_CONNECT_TIMEOUT,
    TRANSFER_TIMEOUT,
)
from ..exceptions import (
    ConnectionFailedError,
    ConnectionReadError,
    ConnectionWriteError,
    MessageDeserializationError,
    MessageSerializationError,
)
from ..protocol import obfuscation
from ..protocol.primitives import uint32, uint64, MessageDataclass
from ..protocol.messages import (
    DistributedMessage,
    PeerInitializationMessage,
    PeerMessage,
    ServerMessage,
)
from ..utils import task_counter

if TYPE_CHECKING:
    from .rate_limiter import RateLimiter
    from .network import Network


logger = logging.getLogger(__name__)

DEFAULT_RECV_BUF_SIZE = 8
"""Default amount of bytes to recv from the socket"""
TRANSFER_RECV_BUF_SIZE = 1024 * 8
"""Default amount of bytes to recv during file transfer"""
TRANSFER_SEND_BUF_SIZE = 1024 * 8
"""Default amount of bytes to send during file transfer"""
HEADER_SIZE_OBFUSCATED: int = obfuscation.KEY_SIZE + struct.calcsize('I')
HEADER_SIZE_UNOBFUSCATED: int = struct.calcsize('I')


class PeerConnectionType:
    FILE = 'F'
    PEER = 'P'
    DISTRIBUTED = 'D'


class ConnectionState(Enum):
    UNINITIALIZED = auto()
    CONNECTING = auto()
    CONNECTED = auto()
    CLOSING = auto()
    CLOSED = auto()


class CloseReason(Enum):
    UNKNOWN = auto()
    CONNECT_FAILED = auto()
    REQUESTED = auto()
    READ_ERROR = auto()
    WRITE_ERROR = auto()
    TIMEOUT = auto()
    EOF = auto()


class PeerConnectionState(Enum):
    AWAITING_INIT = auto()
    """No init message has been received yet (PeerInit, PeerPierceFirewall)"""
    ESTABLISHED = auto()
    """Message connections: ready to receive / send data"""
    AWAITING_TICKET = auto()
    """Transfer connections: awaiting the transfer ticket"""
    AWAITING_OFFSET = auto()
    """Transfer connections: awaiting the transfer offset"""
    TRANSFERRING = auto()
    """Transfer connections: ready to receive / send data"""


class Connection:

    def __init__(self, hostname: str, port: int, network: Network):
        self.hostname: str = hostname
        self.port: int = port
        self.network: Network = network
        self.state: ConnectionState = ConnectionState.UNINITIALIZED

    async def set_state(self, state: ConnectionState, close_reason: CloseReason = CloseReason.UNKNOWN):
        self.state = state
        await self.network.on_state_changed(state, self, close_reason=close_reason)

    def _is_closing(self) -> bool:
        return self.state in (ConnectionState.CLOSING, ConnectionState.CLOSED)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(hostname={self.hostname!r}, port={self.port}, "
            f"state={self.state})")


class ListeningConnection(Connection):
    """A listening connection, objects of this class are responsible for
    accepting incoming connections from peers
    """

    def __init__(self, hostname: str, port: int, network: Network, obfuscated: bool = False):
        super().__init__(hostname, port, network)
        self.obfuscated: bool = obfuscated

        self._server: Optional[asyncio.AbstractServer] = None
        self.connections_accepted: int = 0

    async def disconnect(self, reason: CloseReason = CloseReason.UNKNOWN):
        logger.debug(f"{self.hostname}:{self.port} : disconnecting : {reason.name}")
        await self.set_state(ConnectionState.CLOSING, close_reason=reason)
        try:
            if self._server is not None:
                if self._server.is_serving():
                    self._server.close()
                await self._server.wait_closed()

        except Exception as exc:
            logger.warning(
                f"{self.hostname}:{self.port} : exception while disconnecting", exc_info=exc)

        finally:
            self._server = None
            logger.debug(f"{self.hostname}:{self.port} : disconnected : {reason.name}")
            await self.set_state(ConnectionState.CLOSED, close_reason=reason)

    async def connect(self):
        """Open a listening connection on the current `hostname` and `port`

        :raise ConnectionFailedError: raised when binding failed
        """
        logger.info(
            f"open {self.hostname}:{self.port} : listening connection")
        await self.set_state(ConnectionState.CONNECTING)

        try:
            self._server = await asyncio.start_server(
                self.accept,
                self.hostname,
                self.port,
                family=socket.AF_INET,
                start_serving=True
            )

        except OSError:
            await self.disconnect(CloseReason.CONNECT_FAILED)
            raise ConnectionFailedError(f"{self.hostname}:{self.port} : failed to connect")

        await self.set_state(ConnectionState.CONNECTED)

    async def accept(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        hostname, port = writer.get_extra_info('peername')
        logger.debug(f"{self.hostname}:{self.port} : accepted connection {hostname}:{port}")
        connection = PeerConnection(
            hostname, port, self.network,
            obfuscated=self.obfuscated,
            connection_type=PeerConnectionType.PEER,
            incoming=True
        )
        connection._reader, connection._writer = reader, writer
        await self.network.on_peer_accepted(connection)
        await connection.set_state(ConnectionState.CONNECTED)


class DataConnection(Connection):
    """Connection for message and data transfer"""

    def __init__(self, hostname: str, port: int, network: Network, obfuscated: bool = False):
        super().__init__(hostname, port, network)
        self.obfuscated = obfuscated

        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None
        self._reader_task: Optional[asyncio.Task] = None
        self.read_timeout: Optional[float] = None
        self._queued_messages: List[asyncio.Task] = []

    def get_connecting_ip(self) -> str:
        """Gets the IP address being used to connect to the server/peer.

        The connection needs to be established for this method to work
        """
        return self._writer.get_extra_info('sockname')[0]

    async def connect(self, timeout: float = 30):
        """Opens a TCP connection the hostname:ip provided in this object. The
        state of the connection will be changed to CONNECTING before the attempt
        is made

        Upon success the `state` will be set to CONNECTED and the message
        reader loop will be started. Upon failure the state will be set to
        CLOSED with CONNECT_FAILED as its reason

        :param timeout: timeout in seconds before giving up (default: 30)
        :raise ConnectionFailedError: raised when connection failed or timed out
        """
        logger.info(f"{self.hostname}:{self.port} : connecting")
        await self.set_state(ConnectionState.CONNECTING)

        try:
            self._reader, self._writer = await asyncio.wait_for(
                asyncio.open_connection(self.hostname, self.port), timeout)

        except (Exception, asyncio.TimeoutError) as exc:
            await self.disconnect(CloseReason.CONNECT_FAILED)
            raise ConnectionFailedError(f"{self.hostname}:{self.port} : failed to connect") from exc

        else:
            logger.debug(f"{self.hostname}:{self.port} : connected")
            await self.set_state(ConnectionState.CONNECTED)

    async def disconnect(self, reason: CloseReason = CloseReason.UNKNOWN):
        """Disconnects the TCP connection. The method will not raise an
        exception in case the connection wasn't yet connected

        :param reason: optional reason for the disconnect. This parameter is
            purely informational
        """
        # Prevents this method being called twice during the process of
        # disconnecting. This is necessary because during disconnecting the
        # reader task throw an EOF and attempt to call this method again.
        if self.state in (ConnectionState.CLOSED, ConnectionState.CLOSING):
            return

        await self.set_state(ConnectionState.CLOSING, close_reason=reason)
        logger.debug(f"{self.hostname}:{self.port} : disconnecting : {reason.name}")
        self._cancel_queued_messages()
        try:
            if self._writer is not None:
                if not self._writer.is_closing():
                    self._writer.close()
                await self._writer.wait_closed()

        except Exception as exc:
            logger.warning(f"{self.hostname}:{self.port} : exception while disconnecting : {exc}")

        finally:
            await self.set_state(ConnectionState.CLOSED, close_reason=reason)
            # Because disconnect can be called when read failed setting the
            # reader task to none should be done last
            self._reader_task = None
            self._reader = None
            self._writer = None

    def start_reader_task(self):
        self._reader_task = asyncio.create_task(self._message_reader_loop())

    def stop_reader_task(self):
        if self._reader_task is not None:
            self._reader_task.cancel()
            self._reader_task = None

    # Read/write methods

    async def _read_until_eof(self):
        await self._reader.read(-1)

    async def receive_until_eof(self, raise_exception=True):
        try:
            return await self._read(self._read_until_eof)
        except ConnectionReadError:
            if raise_exception:
                raise

    async def _message_reader_loop(self):
        """Message reader loop. This will loop until the connection is closed or
        the network is closed
        """
        while not self._is_closing():
            try:
                message_data = await self.receive_message()
                if message_data:
                    message = self.decode_message_data(message_data)

            except ConnectionReadError:
                logger.warning(f"{self.hostname}:{self.port} : read error")

            except MessageDeserializationError:
                logger.warning(f"{self.hostname}:{self.port} : failed to deserialize message")

            else:
                # Do not handle messages when closing/closed
                if not self._is_closing():
                    if message:
                        await self._perform_message_callback(message)

    async def _read_message(self):
        header_size = HEADER_SIZE_OBFUSCATED if self.obfuscated else HEADER_SIZE_UNOBFUSCATED
        header = await asyncio.wait_for(
            self._reader.readexactly(header_size), self.read_timeout)

        message_len_buf = obfuscation.decode(header) if self.obfuscated else header
        _, message_len = uint32.deserialize(0, message_len_buf)

        message = await asyncio.wait_for(
            self._reader.readexactly(message_len), self.read_timeout)

        return header + message

    async def receive_message_object(self) -> Optional[MessageDataclass]:
        message_data = await self.receive_message()
        if message_data:
            return self.decode_message_data(message_data)
        else:
            return None

    async def receive_message(self) -> Optional[bytes]:
        """Receives a single raw message"""
        return await self._read(self._read_message)

    def decode_message_data(self, data: bytes) -> MessageDataclass:
        """De-obfuscates and deserializes message data into a `MessageDataclass`
        object. See `deserialize_message`

        :param data: message bytes to decode
        :return: object of `MessageDataclass`
        :raise MessageDeserializationError: raised when deserialization failed
        """
        if self.obfuscated:
            data = obfuscation.decode(data)

        try:
            message = self.deserialize_message(data)
        except Exception as exc:
            logger.exception(f"{self.hostname}:{self.port} : failed to deserialize message : {data.hex()}")
            raise MessageDeserializationError("failed to deserialize message") from exc

        return message

    def encode_message_data(self, message: Union[bytes, MessageDataclass]) -> bytes:
        """Serializes the `MessageDataclass` or `bytes` and obfuscates the
        contents. See `serialize_message`

        :return: `bytes` object
        :raise MessageSerializationError: raised when serialization failed
        """
        try:
            data = self.serialize_message(message)
        except Exception as exc:
            logger.exception(f"{self.hostname}:{self.port} : failed to serialize message : {message!r}")
            raise MessageSerializationError("failed to serialize Message") from exc

        if self.obfuscated:
            data = obfuscation.encode(data)

        return data

    async def _perform_message_callback(self, message: MessageDataclass):
        logger.debug(f"{self.hostname}:{self.port} : received message : {message!r}")
        try:
            await self.network.on_message_received(message, self)
        except Exception:
            logger.exception(f"error during callback : {message!r}")

    async def _read(self, reader_func, timeout: Optional[float] = None) -> Optional[bytes]:
        """Read data from the connection using the passed `reader_func`. When an
        error occurs during reading the connection will be CLOSED

        :param reader_func: callable that reads the data
        :return: the return value of the `reader_func`, None if EOF
        :raise ConnectionReadError: upon any kind of read error/timeout
        """
        try:
            if timeout:
                return await asyncio.wait_for(reader_func(), timeout)
            else:
                return await reader_func()

        except asyncio.IncompleteReadError as exc:
            if exc.partial:
                await self.disconnect(CloseReason.READ_ERROR)
                raise ConnectionReadError(f"{self.hostname}:{self.port} : incomplete read on connection : {exc.partial!r}") from exc
            else:
                await self.disconnect(CloseReason.EOF)
                return None

        except asyncio.TimeoutError as exc:
            await self.disconnect(CloseReason.TIMEOUT)
            raise ConnectionReadError(f"{self.hostname}:{self.port} : read timeout") from exc

        except Exception as exc:
            await self.disconnect(CloseReason.READ_ERROR)
            raise ConnectionReadError(f"{self.hostname}:{self.port} : exception during reading") from exc

    def queue_message(self, message: Union[bytes, MessageDataclass]) -> asyncio.Task:
        task = asyncio.create_task(
            self.send_message(message),
            name=f'queue-message-task-{task_counter()}'
        )
        self._queued_messages.append(task)
        task.add_done_callback(self._queued_messages.remove)
        return task

    def queue_messages(self, *messages: Union[bytes, MessageDataclass]) -> List[asyncio.Task]:
        return [
            self.queue_message(message)
            for message in messages
        ]

    def _cancel_queued_messages(self):
        for qmessage_task in self._queued_messages:
            qmessage_task.cancel()

    async def send_message(self, message: Union[bytes, MessageDataclass]):
        """Sends a message or a set of bytes over the connection. In case an
        object of `MessageDataClass` is provided the object will first be
        serialized. If the `obfuscated` flag is set for the connection the
        message or bytes will first be obfuscated

        :param message: message to be sent over the connection
        :raise ConnectionWriteError: error or timeout occured during writing
        """
        if self._is_closing():
            logger.warning(
                f"{self.hostname}:{self.port} : not sending message, connection is closing / closed : {message!r}")
            return

        logger.debug(f"{self.hostname}:{self.port} : send message : {message!r}")
        # Serialize the message
        try:
            data = self.encode_message_data(message)
        except MessageSerializationError:
            return

        # Perform actual send
        try:
            self._writer.write(data)
            await asyncio.wait_for(self._writer.drain(), 10)

        except asyncio.TimeoutError as exc:
            await self.disconnect(CloseReason.TIMEOUT)
            raise ConnectionWriteError(f"{self.hostname}:{self.port} : write timeout") from exc

        except Exception as exc:
            await self.disconnect(CloseReason.WRITE_ERROR)
            raise ConnectionWriteError(f"{self.hostname}:{self.port} : exception during writing") from exc

    def deserialize_message(self, message_data: bytes) -> MessageDataclass:
        """Should be called after a full message has been received. This method
        should parse the message
        """
        raise NotImplementedError(
            "'deserialize_message' should be overwritten by a subclass")

    def serialize_message(self, message: Union[bytes, MessageDataclass]) -> bytes:
        if isinstance(message, MessageDataclass):
            return message.serialize()
        else:
            return message


class ServerConnection(DataConnection):

    async def connect(self, timeout: float = SERVER_CONNECT_TIMEOUT):
        await super().connect(timeout=timeout)
        # self.start_reader_task()

    def deserialize_message(self, message_data: bytes) -> MessageDataclass:
        return ServerMessage.deserialize_response(message_data)


class PeerConnection(DataConnection):

    def __init__(
            self, hostname: str, port: int, network: Network, obfuscated: bool = False,
            username: Optional[str] = None, connection_type: str = PeerConnectionType.PEER,
            incoming: bool = False):
        super().__init__(hostname, port, network, obfuscated=obfuscated)
        self.incoming: bool = incoming
        self.connection_state = PeerConnectionState.AWAITING_INIT

        self.username: Optional[str] = username
        self.connection_type: str = connection_type
        self.read_timeout = PEER_READ_TIMEOUT

        self.download_rate_limiter: Optional[RateLimiter] = None
        self.upload_rate_limiter: Optional[RateLimiter] = None

    async def connect(self, timeout: float = PEER_CONNECT_TIMEOUT):
        await super().connect(timeout=timeout)

    def set_connection_state(self, state: PeerConnectionState):
        """Sets the current connection state.

        If the current state is AWAITING_INIT and the connection type is a
        distributed or file connection the `obfuscated` flag for this connection
        will be set to `False`

        If the state goes to the ESTABLISHED state the message reader task will
        be started. In all other cases it will be stopped (in case the task was
        running)

        :param state: The new state of the connection
        """
        # Set non-peer connections to non-obfuscated
        if state != PeerConnectionState.AWAITING_INIT:
            if self.connection_type != PeerConnectionType.PEER:
                self.obfuscated = False

        if state == PeerConnectionState.ESTABLISHED:
            self.start_reader_task()
        else:
            # This shouldn't occur, during all other states we call the
            # receive_* methods directly. But it can do no harm
            self.stop_reader_task()

        logger.debug(f"{self.hostname}:{self.port} setting state to {state} : {self!r}")
        self.connection_state = state

    async def _read_transfer_ticket(self) -> bytes:
        return await self._reader.readexactly(struct.calcsize('I'))

    async def _read_transfer_offset(self) -> bytes:
        return await self._reader.readexactly(struct.calcsize('Q'))

    async def receive_transfer_ticket(self) -> int:
        """Receive the transfer ticket from the connection"""
        data = await self._read(self._read_transfer_ticket)
        _, ticket = uint32.deserialize(0, data)
        return ticket

    async def receive_transfer_offset(self) -> int:
        """Receive the transfer offset from the connection"""
        data = await self._read(self._read_transfer_offset)
        _, offset = uint64.deserialize(0, data)
        return offset

    async def receive_data(self, n_bytes: int, timeout: float = TRANSFER_TIMEOUT) -> Optional[bytes]:
        """Receives given amount of bytes on the connection.

        In case of error the socket will be disconnected. If no data is received
        it is assumed to be EOF and the connection will be disconnected.

        :param n_bytes: amount of bytes to receive
        :param timeout: timeout in seconds
        :raise ConnectionReadError: in case timeout occured on an error on the
            socket
        :return: `bytes` object containing the received data
        """
        try:
            data = await asyncio.wait_for(
                self._reader.read(n_bytes), timeout)

        except asyncio.TimeoutError as exc:
            await self.disconnect(CloseReason.TIMEOUT)
            raise ConnectionReadError(f"{self.hostname}:{self.port} : timeout writing") from exc

        except Exception as exc:
            await self.disconnect(CloseReason.READ_ERROR)
            raise ConnectionReadError(f"{self.hostname}:{self.port} : read error") from exc

        else:
            if not data:
                await self.disconnect(CloseReason.EOF)
                return None
            else:
                return data

    async def receive_file(self, file_handle: BinaryIO, filesize: int, callback: Optional[Callable[[bytes], None]] = None):
        """Receives a file on the current connection and writes it to the given
        `file_handle`

        This method will attempt to keep reading data until EOF is received from
        the connection or the amount of received bytes has reached the `filesize`

        :param file_handle: a file handle to write the received data to
        :param filesize: expected filesize
        :param callback: optional callback that gets called each time a chunk
            of data is received
        """
        bytes_received = 0
        while True:
            bytes_to_read = await self.download_rate_limiter.take_tokens()
            data = await self.receive_data(bytes_to_read)
            if data is None:
                return

            await file_handle.write(data)
            if callback is not None:
                callback(data)

            # Check if all data received and return
            bytes_received += len(data)
            if bytes_received >= filesize:
                return

    async def send_data(self, data: bytes):
        try:
            self._writer.write(data)
            await asyncio.wait_for(self._writer.drain(), TRANSFER_TIMEOUT)

        except asyncio.TimeoutError as exc:
            await self.disconnect(CloseReason.TIMEOUT)
            raise ConnectionWriteError(f"{self.hostname}:{self.port} : timeout writing") from exc

        except Exception as exc:
            await self.disconnect(CloseReason.WRITE_ERROR)
            raise ConnectionWriteError(f"{self.hostname}:{self.port} : write error") from exc

    async def send_file(self, file_handle: BinaryIO, callback: Optional[Callable[[bytes], None]] = None):
        """Sends a file over the connection. This method makes use of the
        `upload_rate_limiter` to limit how many bytes are being sent at a time.

        :param file_handle: binary opened file handle
        :param callback: progress callback that gets called each time a chunk
            of data was sent
        """
        while True:
            bytes_to_write = await self.upload_rate_limiter.take_tokens()
            data = await file_handle.read(bytes_to_write)
            if not data:
                return

            await self.send_data(data)
            if callback is not None:
                callback(data)

    def deserialize_message(self, message_data: bytes) -> MessageDataclass:
        if self.connection_state == PeerConnectionState.AWAITING_INIT:
            return PeerInitializationMessage.deserialize_request(message_data)

        else:
            if self.connection_type == PeerConnectionType.PEER:
                return PeerMessage.deserialize_request(message_data)
            else:
                return DistributedMessage.deserialize_request(message_data)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"hostname={self.hostname!r}, port={self.port}, state={self.state}, "
            f"incoming={self.incoming}, obfuscated={self.obfuscated}, username={self.username!r}, "
            f"connection_state={self.connection_state}, connection_type={self.connection_type!r})")



from __future__ import annotations
from dataclasses import dataclass, field
from enum import auto, Enum, Flag
from typing import Optional, Set
from ..protocol.primitives import UserStats


class UserStatus(Enum):
    """User status values, everything except the UNKNOWN status are used by the
    server
    """
    UNKNOWN = -1
    OFFLINE = 0
    AWAY = 1
    ONLINE = 2


class UploadPermissions(Enum):
    """Upload permissions returned / sent by PeerUserInfoReply message"""
    UNKNOWN = -1
    NOONE = 0
    """Don't allow uploads from anyone"""
    EVERYONE = 1
    """Allow uploads from anyone"""
    USER_LIST = 2
    """Only allow uploads from users in the users list"""
    PERMITTED_LIST = 3
    """Only allow uploads from users in a specific list"""


class TrackingFlag(Flag):
    """Tracking flags hold information how the user is being tracked"""
    REQUESTED = auto()
    """Tracking was explicitly requested by the user"""
    TRANSFER = auto()
    """Tracking was requested through transfer manager"""
    FRIEND = auto()
    """Tracking because the user is a friend"""


@dataclass
class User:
    name: str

    # Status info
    status: UserStatus = UserStatus.UNKNOWN
    privileged: bool = False

    # User info
    description: Optional[str] = None
    picture: Optional[bytes] = None
    country: Optional[str] = None

    # User stats
    avg_speed: Optional[int] = None
    uploads: Optional[int] = None
    shared_file_count: Optional[int] = None
    shared_folder_count: Optional[int] = None

    # Upload status
    has_slots_free: Optional[bool] = None
    slots_free: Optional[int] = None
    upload_slots: Optional[int] = None
    queue_length: Optional[int] = None
    upload_permissions: UploadPermissions = UploadPermissions.UNKNOWN

    # Interests
    interests: Set[str] = field(default_factory=set)
    hated_interests: Set[str] = field(default_factory=set)

    def update_from_user_stats(self, user_stats: UserStats):
        self.avg_speed = user_stats.avg_speed
        self.shared_file_count = user_stats.shared_file_count
        self.shared_folder_count = user_stats.shared_folder_count
        self.uploads = user_stats.uploads

    def clear_all(self):
        """Clears all user data except user status and privileges info and
        country
        """
        self.clear(info=True, interests=True, stats=True, upload_info=True)

    def clear(
            self, info: bool = False, interests: bool = False,
            stats: bool = False, upload_info: bool = False):
        """Resets selected fields"""

        if info:
            self.picture = None
            self.description = None

        if interests:
            self.interests = set()
            self.hated_interests = set()

        if stats:
            self.avg_speed = None
            self.has_slots_free = None
            self.slots_free = None
            self.queue_length = None
            self.uploads = None

        if upload_info:
            self.upload_permissions = None
            self.upload_slots = None
            self.has_slots_free = None
            self.queue_length = None
            self.slots_free = None


@dataclass
class ChatMessage:
    id: int
    timestamp: int
    user: User
    message: str
    is_admin: bool

    def is_server_message(self) -> bool:
        return self.is_admin and self.user.name == 'server'



from __future__ import annotations
import asyncio
import enum
from functools import partial
import logging
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
    Tuple,
    TypeVar,
    Type,
    TYPE_CHECKING
)

from ..constants import (
    DEFAULT_LISTENING_HOST,
    PEER_INDIRECT_CONNECT_TIMEOUT,
)
from .connection import (
    Connection,
    ConnectionState,
    CloseReason,
    DataConnection,
    PeerConnectionState,
    ListeningConnection,
    PeerConnection,
    PeerConnectionType,
    ServerConnection,
)
from ..exceptions import (
    ConnectionFailedError,
    ConnectionReadError,
    ListeningConnectionFailedError,
    MessageDeserializationError,
    NetworkError,
    PeerConnectionError,
)
from ..events import (
    build_message_map,
    on_message,
    ConnectionStateChangedEvent,
    EventBus,
    PeerInitializedEvent,
    MessageReceivedEvent,
    SessionInitializedEvent,
    ServerReconnectedEvent,
)
from ..protocol.messages import (
    CannotConnect,
    ConnectToPeer,
    GetPeerAddress,
    MessageDataclass,
    PeerInit,
    PeerPierceFirewall,
    SetListenPort,
)
from . import upnp
from .rate_limiter import RateLimiter
from ..utils import task_counter, ticket_generator

if TYPE_CHECKING:
    from ..settings import Settings


logger = logging.getLogger(__name__)


T = TypeVar('T', bound='MessageDataclass')
ListeningConnections = Tuple[Optional[ListeningConnection], Optional[ListeningConnection]]


class ListeningConnectionErrorMode(enum.Enum):
    """Error mode for listening connections. During initialization of the
    network the `connect_listening_connections` method will raise an error
    depending on the error mode
    """
    ALL = 'all'
    """Raise an exception if all listening connections failed to connect"""
    ANY = 'any'
    """Raise an exception if any listening connections failed to connect"""
    CLEAR = 'clear'
    """Raise an exception only if the non-obfuscated connection failed to
    connect
    """


class ExpectedResponse(asyncio.Future):
    """Future for an expected response message"""

    def __init__(
            self, connection_class: Type[Union[PeerConnection, ServerConnection]], message_class: Type[MessageDataclass],
            peer: Optional[str] = None, fields: Optional[Dict[str, Any]] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None):
        super().__init__(loop=loop)
        self.connection_class: Type[Union[PeerConnection, ServerConnection]] = connection_class
        self.message_class: Type[MessageDataclass] = message_class
        self.peer: Optional[str] = peer
        self.fields: Dict[str, Any] = {} if fields is None else fields

    def matches(self, connection: Union[PeerConnection, ServerConnection], response: MessageDataclass) -> bool:
        if connection.__class__ != self.connection_class:
            return False

        if response.__class__ != self.message_class:
            return False

        if self.peer is not None and isinstance(connection, PeerConnection):
            if connection.username != self.peer:
                return False

        for fname, expected_value in self.fields.items():
            if callable(expected_value):
                try:
                    actual_value = getattr(response, fname)
                except AttributeError:
                    return False
                else:
                    return expected_value(actual_value)

            if getattr(response, fname, None) != expected_value:
                return False

        return True


class Network:

    def __init__(self, settings: Settings, event_bus: EventBus):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._upnp = upnp.UPNP(self._settings)
        self._ticket_generator = ticket_generator()
        self._expected_response_futures: List[ExpectedResponse] = []
        self._expected_connection_futures: Dict[int, asyncio.Future] = {}

        # List of connections
        self.server_connection: ServerConnection = self.create_server_connection()
        self.listening_connections: ListeningConnections = self.create_listening_connections()
        self.peer_connections: List[PeerConnection] = []

        self._MESSAGE_MAP = build_message_map(self)

        # Rate limiters
        self._upload_rate_limiter: RateLimiter = RateLimiter.create_limiter(
            self._settings.network.limits.upload_speed_kbps)
        self._download_rate_limiter: RateLimiter = RateLimiter.create_limiter(
            self._settings.network.limits.download_speed_kbps)

        # Debugging
        self._ip_overrides: Dict[str, str] = self._settings.debug.ip_overrides

        # Operational
        self._log_connections_task: Optional[asyncio.Task] = None
        self._create_peer_connection_tasks: List[asyncio.Task] = []
        self._upnp_task: Optional[asyncio.Task] = None
        self._connection_watchdog_task: Optional[asyncio.Task] = None

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)

    def create_server_connection(self) -> ServerConnection:
        return ServerConnection(
            self._settings.network.server.hostname,
            self._settings.network.server.port,
            self
        )

    def create_listening_connections(self) -> ListeningConnections:
        return (
            ListeningConnection(
                DEFAULT_LISTENING_HOST,
                self._settings.network.listening.port,
                self,
                obfuscated=False
            ) if self._settings.network.listening.port else None,
            ListeningConnection(
                DEFAULT_LISTENING_HOST,
                self._settings.network.listening.obfuscated_port,
                self,
                obfuscated=True
            ) if self._settings.network.listening.obfuscated_port else None
        )

    async def initialize(self):
        """Initializes the server and listening connections"""
        logger.info("initializing network")

        await self.connect_listening_ports()
        await self.connect_server()

        log_connections = self._settings.debug.log_connection_count
        if log_connections and self._log_connections_task is None:
            self._log_connections_task = asyncio.create_task(
                self._log_connections_job(),
                name=f'log-connections-{task_counter()}'
            )

    async def connect_listening_ports(self):
        """This method will attempt to connect both listening ports (if
        configured) but will raise an exception depending on the settings
        `network.listening.error_mode` setting. All other listening connections
        will be disconnected before raising the error

        :raise ListeningConnectionFailedError: if an error occurred connecting
            the listening ports
        """
        error_mode = self._settings.network.listening.error_mode
        results = await asyncio.gather(
            *[
                listening_connection.connect()
                for listening_connection in self.listening_connections
                if listening_connection
            ],
            return_exceptions=True
        )

        if error_mode == ListeningConnectionErrorMode.ALL:
            if all(isinstance(res, ConnectionFailedError) for res in results):
                await self.disconnect_listening_ports()
                raise ListeningConnectionFailedError("failed to open any listening ports")

        elif error_mode == ListeningConnectionErrorMode.ANY:
            if any(isinstance(res, ConnectionFailedError) for res in results):
                await self.disconnect_listening_ports()
                raise ListeningConnectionFailedError("one or more listening ports failed to connect")

        elif error_mode == ListeningConnectionErrorMode.CLEAR:
            if not self.listening_connections[0] or self.listening_connections[0].state != ConnectionState.CONNECTED:
                await self.disconnect_listening_ports()
                raise ListeningConnectionFailedError("failed to connect non-obfuscated listening port")

    async def disconnect_listening_ports(self):
        """Disconnects all listening ports"""
        await asyncio.gather(
            *[
                listening_connection.disconnect(CloseReason.REQUESTED)
                for listening_connection in self.listening_connections
                if listening_connection
            ],
            return_exceptions=True
        )

    async def connect_server(self):
        await self.server_connection.connect()

    async def disconnect_server(self):
        await self.server_connection.disconnect(CloseReason.REQUESTED)

    async def disconnect(self):
        """Cancels all connection creation tasks and disconnects all current
        open connections
        """
        self._cancel_all_tasks()

        connections = [self.server_connection, ] + self.peer_connections
        connections += [conn for conn in self.listening_connections if conn]
        logger.info(f"waiting for network disconnect : {len(connections)} connections")
        await asyncio.gather(
            *[conn.disconnect(CloseReason.REQUESTED) for conn in connections],
            return_exceptions=True
        )
        logger.info("network disconnected")

    def get_listening_ports(self) -> Tuple[int, int]:
        """Gets the currently connected listening ports

        :return: `tuple` with 2 elements: the non-obfuscated- and obfuscated
            ports. If either is not connected or not configured `0` will be
            returned for this port
        """
        def get_port(conn: Optional[ListeningConnection]):
            if conn and conn.state == ConnectionState.CONNECTED:
                return conn.port
            return 0

        nonobf_port = get_port(self.listening_connections[0])
        obf_port = get_port(self.listening_connections[1])
        return nonobf_port, obf_port

    async def advertise_listening_ports(self):
        """Notifies the server of our current listening ports"""
        port, obfuscated_port = self.get_listening_ports()

        await self.send_server_messages(
            SetListenPort.Request(
                port,
                obfuscated_port_amount=1 if obfuscated_port else 0,
                obfuscated_port=obfuscated_port
            )
        )

    # Settings listeners
    def load_speed_limits(self):
        """(Re)loads the speed limits from the settings"""
        self.set_download_speed_limit(
            self._settings.network.limits.download_speed_kbps)
        self.set_upload_speed_limit(
            self._settings.network.limits.upload_speed_kbps)

    def set_upload_speed_limit(self, limit_kbps: int):
        """Modifies the upload speed limit. Passing 0 will set the upload speed
        to unlimited

        :param limit_kbps: the new upload limit
        """
        new_limiter = RateLimiter.create_limiter(limit_kbps)
        if self._upload_rate_limiter is not None:
            new_limiter.copy_tokens(self._upload_rate_limiter)
        self._upload_rate_limiter = new_limiter

        for conn in self.peer_connections:
            conn.upload_rate_limiter = self._upload_rate_limiter

    def set_download_speed_limit(self, limit_kbps: int):
        """Modifies the download speed limit. Passing 0 will set the download
        speed to unlimited

        :param limit_kbps: the new download limit
        """
        new_limiter = RateLimiter.create_limiter(limit_kbps)
        if self._download_rate_limiter is not None:
            new_limiter.copy_tokens(self._download_rate_limiter)
        self._download_rate_limiter = new_limiter

        for conn in self.peer_connections:
            conn.download_rate_limiter = self._download_rate_limiter

    async def _server_connection_watchdog_job(self):
        """Reconnects to the server if it is closed. This should be started as a
        task and should be cancelled upon request.
        """
        timeout = self._settings.network.server.reconnect.timeout
        logger.info("starting server connection watchdog")
        last_state = self.server_connection.state
        while True:
            await asyncio.sleep(0.5)

            if self.server_connection.state == ConnectionState.CLOSED:

                if not self._settings.credentials.are_configured():
                    if self.server_connection.state != last_state:
                        logger.warning(
                            "credentials are not correctly configured in the "
                            "settings, will not attempt to reconnect"
                        )
                    continue

                else:
                    logger.info(f"will attempt to reconnect to server in {timeout} seconds")
                    await asyncio.sleep(timeout)
                    try:
                        await self.connect_server()
                    except ConnectionFailedError:
                        logger.warning("failed to reconnect to server")
                    else:
                        await self._event_bus.emit(
                            ServerReconnectedEvent())

            last_state = self.server_connection.state

    async def start_server_connection_watchdog(self):
        """Starts the server connection watchdog if it is not already running"""
        if self._connection_watchdog_task is None:
            self._connection_watchdog_task = asyncio.create_task(
                self._server_connection_watchdog_job(),
                name=f'watchdog-task-{task_counter()}'
            )

    def stop_server_connection_watchdog(self):
        """Stops the server connection watchdog if it is running"""
        if self._connection_watchdog_task is not None:
            self._connection_watchdog_task.cancel()
            self._connection_watchdog_task = None

    async def _log_connections_job(self):
        while True:
            await asyncio.sleep(10)
            count = len(self.peer_connections)
            logger.info(
                f"currently {count} peer connections ({len(self._create_peer_connection_tasks)} tasks)")

    async def map_upnp_ports(self):
        """Maps the listening ports using UPNP on the gateway"""
        ports = [port for port in self.get_listening_ports() if port]
        ip_address = self.server_connection.get_connecting_ip()

        devices = await self._upnp.search_igd_devices(ip_address)
        for port in ports:
            for device in devices:
                await self._upnp.map_port(device, ip_address, port)

    def _map_upnp_ports_callback(self, task: asyncio.Task):
        try:
            task.result()
        except asyncio.CancelledError:
            logger.warning("upnp: port mapping task cancelled")
        except Exception:
            logger.exception("upnp: failed to map ports")
        finally:
            self._upnp_task = None

    def select_port(self, port, obfuscated_port) -> Tuple[int, bool]:
        """Selects the port used for making a connection. This attempts to take
        into account the `network.peer.obfuscate` parameter however falls back
        on using whatever port is available if necessary

        :param port: available non-obfuscated port
        :param obfuscated_port: available obfuscated port
        :return: a tuple with the port number and whether the port is obfuscated
        """
        prefer_obfuscated = self._settings.network.peer.obfuscate
        if port and obfuscated_port:
            if prefer_obfuscated:
                return obfuscated_port, True
            else:
                return port, False
        elif port:
            return port, False
        else:
            return obfuscated_port, True

    async def create_peer_connection(
            self, username: str, typ: str, ip: Optional[str] = None, port: Optional[int] = None,
            obfuscate: bool = False,
            initial_state: PeerConnectionState = PeerConnectionState.ESTABLISHED) -> PeerConnection:
        """Creates a new peer connection to the given `username` and connection
        type.

        Optionally this method takes `ip`, `port` and `obfuscate` parameters. If
        not given it will be requested to the server (GetPeerAddress) before
        attempting connection

        :param username: username of the peer
        :param typ: connection type ('P', 'D', 'F')
        :param ip: optional ip address of the peer
        :param port: optional port of the peer
        :param obfuscate: whether to obfuscate the connection. Should be used in
            combination with the `ip` and `port`
        :return: an object of `PeerConnection`
        :raise PeerConnectionError: when establishing a peer connection has
            failed
        """
        ticket = next(self._ticket_generator)

        # Request peer address if ip and port are not given
        if ip is None or port is None:
            ip, clear_port, obfuscated_port = await self._get_peer_address(username)

            port, obfuscate = self.select_port(clear_port, obfuscated_port)

        # Override IP address if requested
        ip = self._ip_overrides.get(username, ip)

        # Make direct connection
        try:
            connection = await self._make_direct_connection(
                ticket, username, typ, ip, port, obfuscate
            )

        except NetworkError as exc:
            logger.debug(
                f"direct connection ({typ}) to peer failed : {username} {ip}:{port} : {exc!r}")

            # Make indirect connection
            try:
                connection = await self._make_indirect_connection(
                    ticket, username, typ
                )

            except NetworkError as exc:
                logger.debug(
                    f"indirect connection to peer failed : {username} : {exc!r}")
                raise PeerConnectionError(
                    f"failed to connect to peer {username} ({typ=}, {ticket=})")

        connection.set_connection_state(initial_state)
        if typ == PeerConnectionType.FILE:
            connection.download_rate_limiter = self._download_rate_limiter
            connection.upload_rate_limiter = self._upload_rate_limiter

        await self._event_bus.emit(
            PeerInitializedEvent(connection, requested=True))
        return connection

    async def _get_peer_address(self, username: str) -> Tuple[str, int, int]:
        """Requests the peer address for the given `username` from the server.

        :raise PeerConnectionError: if no IP address or no valid ports were
            returned
        """
        await self.server_connection.send_message(GetPeerAddress.Request(username))
        _, response = await self.create_server_response_future(
            GetPeerAddress.Response,
            fields={
                'username': username
            }
        )

        if response.ip == '0.0.0.0':
            logger.warning(f"GetPeerAddress : no address returned for username : {username}")
            raise PeerConnectionError(f"no address for user : {username}")

        elif not response.port and not response.obfuscated_port:
            logger.warning(f"GetPeerAddress : no valid ports found for user : {username}")
            raise PeerConnectionError(f"no valid ports for user : {username}")

        return response.ip, response.port, response.obfuscated_port

    async def get_peer_connection(self, username: str, typ: str = PeerConnectionType.PEER) -> PeerConnection:
        """Gets a peer connection for the given `username`. It will first try to
        re-use an existing connection, otherwise it will create a new connection

        :param username: username of the peer to connect to
        :param typ: Type of connection to create or reuse
        :return: A `PeerConnection` instance
        """
        connections = self.get_active_peer_connections(username, typ)
        if connections:
            return connections[0]

        return await self.create_peer_connection(username, typ)

    def _remove_response_future(self, expected_response: ExpectedResponse):
        if expected_response in self._expected_response_futures:
            self._expected_response_futures.remove(expected_response)

    def _remove_connection_future(self, ticket: int, connection_future: asyncio.Future):
        logger.debug(f"removing expected incoming connection with {ticket=}")
        self._expected_connection_futures.pop(ticket)

    def create_server_response_future(
            self, message_class: Type[T], fields: Optional[Dict[str, Any]] = None) -> ExpectedResponse:
        """Creates a future for a server message to arrive, the message must
        match the `message_class` and fields defined in the keyword arguments.

        The future will be stored by the current class and will be removed once
        the future is complete or cancelled.

        :param message_class: The expected `MessageDataClass`
        :return: `ExpectedResponse` object
        """
        fields = fields or {}
        future = ExpectedResponse(
            ServerConnection,
            message_class=message_class,
            fields=fields
        )
        self._expected_response_futures.append(future)
        future.add_done_callback(self._remove_response_future)
        return future

    def register_response_future(self, expected_response: ExpectedResponse):
        """Registers an expected response future"""
        self._expected_response_futures.append(expected_response)
        expected_response.add_done_callback(self._remove_response_future)

    async def wait_for_server_message(
            self, message_class: Type[T], fields: Optional[Dict[str, Any]] = None, timeout: float = 10) -> T:
        """Waits for a message from the server

        :param message_class: Class of the expected server message
        :param fields: Optional matchers for the message fields
        :return: The `MessageData` object corresponding to the response
        """
        future = self.create_server_response_future(
            message_class=message_class,
            fields=fields
        )
        try:
            _, response = await asyncio.wait_for(future, timeout=timeout)
        except TimeoutError as exc:
            future.set_exception(exc)
            raise

        return response

    def create_peer_response_future(
            self, peer: str, message_class: Type[T], fields: Optional[Dict[str, Any]] = None) -> ExpectedResponse:
        """Creates a future for a peer message to arrive, the message must match
        the `message_class` and fields defined in the keyword arguments and
        must be coming from a connection by `peer`.

        The future will be stored by the current class and will be removed once
        the future is complete or cancelled.

        :param peer: name of the peer for which the message should be received
        :return: `ExpectedResponse` object
        """
        fields = fields or {}
        future = ExpectedResponse(
            PeerConnection,
            message_class=message_class,
            peer=peer,
            fields=fields
        )
        self._expected_response_futures.append(future)
        future.add_done_callback(self._remove_response_future)
        return future

    async def wait_for_peer_message(
            self, peer: str, message_class: Type[T], fields: Optional[Dict[str, Any]] = None, timeout: float = 60) -> T:
        future = self.create_peer_response_future(
            peer=peer,
            message_class=message_class,
            fields=fields
        )
        try:
            _, response = await asyncio.wait_for(future, timeout=timeout)
        except TimeoutError as exc:
            future.set_exception(exc)
            raise

        return response

    def get_peer_connections(self, username: str, typ: str) -> List[PeerConnection]:
        """Returns all connections for peer with given username and peer
        connection types.

        :param username: username of the peer
        :param typ: type of the connection (from `PeerConnectionType`)
        :return: list of connections for that matches the parameters
        """
        return [
            connection for connection in self.peer_connections
            if connection.username == username and connection.connection_type == typ
        ]

    def get_active_peer_connections(
            self, username: str, typ: str) -> List[PeerConnection]:
        """Return a list of currently active messaging connections for given
        peer connection type.

        :param username: username for which to get the active peer
        :param typ: peer connection type
        :return: list of PeerConnection instances
        """
        return list(
            filter(
                lambda conn: conn.state == ConnectionState.CONNECTED and conn.connection_state == PeerConnectionState.ESTABLISHED,
                self.get_peer_connections(username, typ)
            )
        )

    def remove_peer_connection(self, connection: PeerConnection):
        if connection in self.peer_connections:
            self.peer_connections.remove(connection)

    # Server related

    @on_message(ConnectToPeer.Response)
    async def _on_connect_to_peer(self, message: ConnectToPeer.Response, connection: ServerConnection):
        """Handles a ConnectToPeer request coming from another peer"""
        task = asyncio.create_task(
            self._handle_connect_to_peer(message),
            name=f'connect-to-peer-{task_counter()}'
        )
        task.add_done_callback(
            partial(self._handle_connect_to_peer_callback, message)
        )
        self._create_peer_connection_tasks.append(task)

    async def _make_direct_connection(
            self, ticket: int, username: str, typ: str, ip: str, port: int, obfuscate: bool) -> PeerConnection:
        """Attempts to make a direct connection to the peer and send a `PeerInit`
        message. This will be the first step in case we are the one initiating
        the connection
        """
        logger.debug(f"attempting to connect to peer : {username!r} {ip}:{port}")
        connection = PeerConnection(
            ip, port, self,
            connection_type=typ,
            username=username,
            obfuscated=obfuscate
        )
        self.peer_connections.append(connection)

        await connection.connect()
        await connection.send_message(
            PeerInit.Request(
                self._settings.credentials.username,
                typ,
                ticket
            )
        )
        return connection

    async def _make_indirect_connection(
            self, ticket: int, username: str, typ: str) -> PeerConnection:
        """Attempts to make an indirect connection by sending a `ConnectToPeer`
        message to the server. The method will wait for an incoming connection
        with a `PeerPierceFirewall` message containing the provided ticket or a
        `CannotConnect` message from the server.

        :return: `PeerConnection` object in case of success
        :raise PeerConnectionError: in case timeout was reached or we received a
            `CannotConnect` message from the server containing the `ticket`
        """
        # Wait for either a established connection with PeerPierceFirewall or a
        # CannotConnect from the server
        expected_connection_future: asyncio.Future = asyncio.Future()
        expected_connection_future.add_done_callback(
            partial(self._remove_connection_future, ticket)
        )
        self._expected_connection_futures[ticket] = expected_connection_future

        cannot_connect_future = self.create_server_response_future(
            CannotConnect.Response,
            fields={
                'ticket': ticket
            }
        )

        # Send the connect to peer message
        await self.server_connection.send_message(
            ConnectToPeer.Request(ticket, username, typ))

        futures = (expected_connection_future, cannot_connect_future)
        done, pending = await asyncio.wait(
            futures,
            timeout=PEER_INDIRECT_CONNECT_TIMEOUT,
            return_when=asyncio.FIRST_COMPLETED
        )

        # Whatever happens here, we can cancel all pending futures
        [fut.cancel() for fut in pending]

        # `done` will be empty in case of timeout
        if not done:
            raise PeerConnectionError(
                f"indirect connection timed out ({username=}, {ticket=})")

        completed_future = done.pop()

        if completed_future == cannot_connect_future:
            raise PeerConnectionError(
                f"indirect connection failed ({username=}, {ticket=})")

        connection = completed_future.result()

        connection.username = username
        connection.connection_type = typ

        return connection

    async def _handle_connect_to_peer(self, message: ConnectToPeer.Response):
        """Handles an indirect connection request received from the server.

        A task is created for this coroutine when a `ConnectToPeer` message is
        received.

        A `PeerInitializedEvent` will be emitted if the connection has been
        successfully established

        :param message: received `ConnectToPeer` message
        """
        ip = self._ip_overrides.get(message.username, message.ip)
        port, obfuscate = self.select_port(message.port, message.obfuscated_port)

        peer_connection = PeerConnection(
            ip, port, self,
            connection_type=message.typ,
            obfuscated=obfuscate,
            username=message.username
        )
        self.peer_connections.append(peer_connection)

        try:
            await peer_connection.connect()
            await peer_connection.send_message(
                PeerPierceFirewall.Request(message.ticket)
            )

        except NetworkError:
            await self.server_connection.queue_message(
                CannotConnect.Request(
                    ticket=message.ticket,
                    username=message.username
                )
            )
            raise PeerConnectionError("failed connect on user request")

        if message.typ == PeerConnectionType.FILE:
            peer_connection.set_connection_state(PeerConnectionState.AWAITING_TICKET)
            peer_connection.download_rate_limiter = self._download_rate_limiter
            peer_connection.upload_rate_limiter = self._upload_rate_limiter
        else:
            peer_connection.set_connection_state(PeerConnectionState.ESTABLISHED)

        await self._event_bus.emit(
            PeerInitializedEvent(peer_connection, requested=False))

    async def send_peer_messages(
            self, username: str, *messages: Union[bytes, MessageDataclass],
            raise_on_error: bool = True):
        """Sends a list of messages to the peer with given `username`. This uses
        `get_peer_connection` and will attempt to re-use a connection or create
        a new peer (P) connection

        :param username: Peer username to send the messages to
        :param messages: List of messages to send
        :param raise_on_error: When `True` an exception is raised when a message
            failed to send, if `False` a list of tuples with the result for each
            message will be returned
        :return: a list of tuples containing the sent messages and the result
            of the sent messages. If `None` is returned for a message it was
            successfully sent, otherwise the result will contain the exception.
            `None` if the `raise_on_error` is `True`
        """
        connection = await self.get_peer_connection(username)
        results = await asyncio.gather(
            *[connection.send_message(message) for message in messages],
            return_exceptions=not raise_on_error
        )
        if not raise_on_error:
            return list(zip(messages, results))

    def queue_server_messages(self, *messages: Union[bytes, MessageDataclass]) -> List[asyncio.Task]:
        """Queues server messages

        :param messages: list of messages to queue
        """
        return self.server_connection.queue_messages(*messages)

    async def send_server_messages(
            self, *messages: Union[bytes, MessageDataclass],
            raise_on_error: bool = True):
        """Sends a list of messages to the server

        :param messages: List of messages to send
        :param raise_on_error: When `True` an exception is raised when a message
            failed to send, if `False` a list of tuples with the result for each
            message will be returned
        :return: a list of tuples containing the sent messages and the result
            of the sent messages. If `None` is returned for a message it was
            successfully sent, otherwise the result will contain the exception.
            `None` if the `raise_on_error` is `True`
        """
        results = await asyncio.gather(
            *[self.server_connection.send_message(message) for message in messages],
            return_exceptions=not raise_on_error
        )
        if not raise_on_error:
            return list(zip(messages, results))

    # Connection state changes
    async def on_state_changed(
            self, state: ConnectionState, connection: Connection,
            close_reason: CloseReason = CloseReason.UNKNOWN):
        """Called when the state of a connection changes. This method calls 3
        private method based on the type of `connection` that was passed

        :param state: the new state the connection has received
        :param connection: the connection for which the state changed
        :param close_reason: in case ConnectionState.CLOSED is passed a reason
            will be given as well
        """
        if isinstance(connection, ServerConnection):
            await self._on_server_connection_state_changed(state, connection, close_reason=close_reason)

        elif isinstance(connection, PeerConnection):
            await self._on_peer_connection_state_changed(state, connection, close_reason=close_reason)

        await self._event_bus.emit(
            ConnectionStateChangedEvent(connection, state, close_reason)
        )

    async def _on_server_connection_state_changed(
            self, state: ConnectionState, connection: ServerConnection,
            close_reason: CloseReason = CloseReason.UNKNOWN):
        if state == ConnectionState.CONNECTED:
            # For registering with UPNP we need to know our own IP first, we can
            # get this from the server connection but we first need to be
            # fully connected to it before we can request a valid IP
            if self._settings.network.upnp.enabled:
                self._upnp_task = asyncio.create_task(
                    self.map_upnp_ports(),
                    name=f'enable-upnp-{task_counter()}'
                )
                self._upnp_task.add_done_callback(self._map_upnp_ports_callback)

            # Register server connection watchdog
            if self._settings.network.server.reconnect.auto:
                await self.start_server_connection_watchdog()

        elif state == ConnectionState.CLOSING:

            # When `disconnect` is called on the connection it will always first
            # go into the CLOSING state. The watchdog will only attempt to
            # reconnect if the server is in CLOSED state. So the code needs to
            # make a decision here whether we want to reconnect or not before it
            # goes into CLOSED state.
            # Cancel the watchdog only if we are closing up on request or the
            # server sends an EOF. The server sends an EOF as soon as you are
            # connected if your IP address is banned, this happens when you make
            # too many connections in a short period of time. Making any more
            # connections will only extend your ban period. Possibly the server
            # will also EOF when the login message is not sent after a certain
            # period of time
            # An EOF will also be sent in case you are kicked due to the same
            # login being used in another location
            if close_reason == CloseReason.REQUESTED:
                self.stop_server_connection_watchdog()

            elif close_reason == CloseReason.EOF:
                logger.warning(
                    "server closed connection, will not attempt to reconnect")
                self.stop_server_connection_watchdog()

    async def _on_peer_connection_state_changed(
            self, state: ConnectionState, connection: PeerConnection,
            close_reason: CloseReason = CloseReason.UNKNOWN):

        if state == ConnectionState.CLOSED:
            self.remove_peer_connection(connection)

    # Peer related
    async def on_peer_accepted(self, connection: PeerConnection):
        """Called when a connection has been accepted on one of the listening
        connections. This method will wait for the peer initialization message:
        either PeerInit or PeerPierceFirewall.

        In case of PeerInit the method will perform the initialization of the
        connection.

        In case of PeerPierceFirewall we look for an associated future object
        for the ticket provided in the message; there should be a task waiting
        for it to be completed (see `_make_indirect_connection`). Full
        initialization of the connection should be done in that method.

        In any other case we disconnect the connection.

        :param connection: the accepted `PeerConnection`
        """
        self.peer_connections.append(connection)

        try:
            message_data = await connection.receive_message()
            if message_data:
                peer_init_message = connection.decode_message_data(message_data)
            else:
                # EOF reached before receiving a message. Connection should've
                # been automatically closed
                return
        except ConnectionReadError:
            # Some read error. Connection should've been automatically closed
            return
        except MessageDeserializationError:
            # Couldn't deserialize the first message. Assume something is broken
            # and disconnect
            await connection.disconnect(CloseReason.REQUESTED)
            return

        if isinstance(peer_init_message, PeerInit.Request):
            connection.username = peer_init_message.username
            connection.connection_type = peer_init_message.typ
            # When the first message is PeerInit it's up to the other peer to send
            # us the transfer ticket in case it's a file connection
            if peer_init_message.typ == PeerConnectionType.FILE:
                connection.set_connection_state(PeerConnectionState.AWAITING_TICKET)
                connection.download_rate_limiter = self._download_rate_limiter
                connection.upload_rate_limiter = self._upload_rate_limiter
            else:
                connection.set_connection_state(PeerConnectionState.ESTABLISHED)

            await self._event_bus.emit(
                PeerInitializedEvent(connection, requested=False))

        elif isinstance(peer_init_message, PeerPierceFirewall.Request):
            ticket = peer_init_message.ticket
            try:
                connection_future = self._expected_connection_futures[ticket]
            except KeyError:
                logger.warning(
                    f"{connection.hostname}:{connection.port} : unknown pierce firewall ticket : {ticket}")
                await connection.disconnect(CloseReason.REQUESTED)
            else:
                connection_future.set_result(connection)

        else:
            logger.warning(
                f"{connection.hostname}:{connection.port} : unknown peer init message : {peer_init_message}")
            await connection.disconnect(CloseReason.REQUESTED)

    async def on_message_received(self, message: MessageDataclass, connection: DataConnection):
        """Method called by `connection` instances when a message is received

        :param message: Received message instance
        :param connection: Connection on which the message was received
        """
        # Call the message callbacks for this object
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, connection)

        # Emit on the message bus. Event handling has priority over completing
        # the future responses
        await self._event_bus.emit(MessageReceivedEvent(message, connection))

        # Complete expected response futures
        for expected_response in self._expected_response_futures:
            if expected_response.matches(connection, message):
                expected_response.set_result((connection, message, ))

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        await self.advertise_listening_ports()

    # Task callbacks
    def _handle_connect_to_peer_callback(self, message: ConnectToPeer.Response, task: asyncio.Task):
        try:
            task.result()
        except asyncio.CancelledError:
            logger.debug(f"cancelled ConnectToPeer request : {message}")
        except Exception as exc:
            logger.warning(f"failed to fulfill ConnectToPeer request : {exc!r} : {message}")
        else:
            logger.info(f"successfully fulfilled ConnectToPeer request : {message}")
        finally:
            self._create_peer_connection_tasks.remove(task)

    def _cancel_all_tasks(self):
        for task in self._create_peer_connection_tasks:
            task.cancel()

        if self._log_connections_task is not None:
            self._log_connections_task.cancel()
            self._log_connections_task = None

        if self._upnp_task is not None:
            self._upnp_task.cancel()
            self._upnp_task = None



import asyncio
from collections import deque
from functools import partial
import logging
from typing import Deque, Dict, List, Optional, Union

from ..base_manager import BaseManager
from ..events import (
    on_message,
    build_message_map,
    EventBus,
    ConnectionStateChangedEvent,
    MessageReceivedEvent,
    SearchRequestReceivedEvent,
    SearchResultEvent,
    SessionDestroyedEvent,
    SessionInitializedEvent,
)
from ..protocol.messages import (
    RoomSearch,
    DistributedSearchRequest,
    DistributedServerSearchRequest,
    FileSearch,
    PeerSearchReply,
    SearchInactivityTimeout,
    ServerSearchRequest,
    UserSearch,
    WishlistInterval,
    WishlistSearch,
)
from ..network.connection import (
    CloseReason,
    ConnectionState,
    PeerConnection,
    ServerConnection,
)
from ..network.network import Network
from ..room.model import Room
from ..settings import Settings
from ..shares.manager import SharesManager
from ..shares.utils import convert_items_to_file_data
from ..session import Session
from ..transfer.interface import UploadInfoProvider
from ..utils import task_counter, ticket_generator
from .model import ReceivedSearch, SearchResult, SearchRequest, SearchType


logger = logging.getLogger(__name__)


class SearchManager(BaseManager):
    """Handler for searches requests"""

    def __init__(
            self, settings: Settings, event_bus: EventBus,
            shares_manager: SharesManager, upload_info_provider: UploadInfoProvider,
            network: Network):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network
        self._shares_manager: SharesManager = shares_manager
        self._upload_info_provider: UploadInfoProvider = upload_info_provider

        self._ticket_generator = ticket_generator()
        self._session: Optional[Session] = None

        self.received_searches: Deque[ReceivedSearch] = deque(list(), 500)
        self.requests: Dict[int, SearchRequest] = {}

        # Server variables
        self.search_inactivity_timeout: Optional[int] = None
        self.wishlist_interval: Optional[int] = None

        self.register_listeners()

        self._MESSAGE_MAP = build_message_map(self)

        self._search_reply_tasks: List[asyncio.Task] = []
        self._wishlist_task: Optional[asyncio.Task] = None

    def register_listeners(self):
        self._event_bus.register(
            ConnectionStateChangedEvent, self._on_state_changed)
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)
        self._event_bus.register(
            SessionDestroyedEvent, self._on_session_destroyed)

    def remove_request(self, request: Union[SearchRequest, int]):
        """Removes the search request from the client. Incoming results after
        the request has been removed will be ignored

        :param request: `SearchRequest` object or ticket number to remove
        """
        ticket = request if isinstance(request, int) else request.ticket
        self.requests.pop(ticket)

    async def search(self, query: str) -> SearchRequest:
        """Performs a global search. The results generated by this query will
        stored in the returned object or can be listened to through the
        `SearchResultEvent` event

        :param query: The search query
        :return: An object containing the search request details and results
        """
        ticket = next(self._ticket_generator)

        await self._network.send_server_messages(
            FileSearch.Request(ticket, query)
        )
        self.requests[ticket] = SearchRequest(
            ticket=ticket,
            query=query,
            search_type=SearchType.NETWORK
        )
        return self.requests[ticket]

    async def search_room(self, room: Union[str, Room], query: str) -> SearchRequest:
        """Performs a search request on the specific user. The results generated
        by this query will stored in the returned object or can be listened to
        through the `SearchResultEvent` event

        :param room: Room object or name to query
        :param query: The search query
        :return: An object containing the search request details and results
        """
        room_name = room.name if isinstance(room, Room) else room
        ticket = next(self._ticket_generator)

        await self._network.send_server_messages(
            RoomSearch.Request(room_name, ticket, query)
        )
        self.requests[ticket] = SearchRequest(
            ticket=ticket,
            query=query,
            search_type=SearchType.ROOM,
            room=room_name
        )
        return self.requests[ticket]

    async def search_user(self, username: str, query: str) -> SearchRequest:
        """Performs a search request on the specific user. The results generated
        by this query will stored in the returned object or can be listened to
        through the `SearchResultEvent` event

        :param user: User object or name to query
        :param query: The search query
        :return: An object containing the search request details and results
        """
        ticket = next(self._ticket_generator)

        await self._network.send_server_messages(
            UserSearch.Request(username, ticket, query)
        )
        self.requests[ticket] = SearchRequest(
            ticket=ticket,
            query=query,
            search_type=SearchType.USER,
            username=username
        )
        return self.requests[ticket]

    async def _query_shares_and_reply(self, ticket: int, username: str, query: str):
        """Performs a query on the shares manager and reports the results to the
        user
        """
        if not self._session:
            logger.warning("not returning search results : no valid session was set")
            return

        visible, locked = self._shares_manager.query(query, username=username)

        result_count = len(visible) + len(locked)
        self.received_searches.append(
            ReceivedSearch(
                username=username,
                query=query,
                result_count=result_count
            )
        )
        await self._event_bus.emit(
            SearchRequestReceivedEvent(
                username=username,
                query=query,
                result_count=result_count
            )
        )

        if len(visible) + len(locked) == 0:
            return

        logger.info(f"found {len(visible)}/{len(locked)} results for query {query!r} (username={username!r})")

        task = asyncio.create_task(
            self._network.send_peer_messages(
                username,
                PeerSearchReply.Request(
                    username=self._session.user.name,
                    ticket=ticket,
                    results=convert_items_to_file_data(visible, use_full_path=True),
                    has_slots_free=self._upload_info_provider.has_slots_free(),
                    avg_speed=int(self._upload_info_provider.get_average_upload_speed()),
                    queue_size=self._upload_info_provider.get_queue_size(),
                    locked_results=convert_items_to_file_data(locked, use_full_path=True)
                )
            ),
            name=f'search-reply-{task_counter()}'
        )
        task.add_done_callback(
            partial(self._search_reply_task_callback, ticket, username, query))
        self._search_reply_tasks.append(task)

    def _search_reply_task_callback(self, ticket: int, username: str, query: str, task: asyncio.Task):
        """Callback for a search reply task. This callback simply logs the
        results and removes the task from the list
        """
        try:
            task.result()

        except asyncio.CancelledError:
            logger.debug(
                f"cancelled delivery of search results (ticket={ticket}, username={username}, query={query})")
        except Exception as exc:
            logger.warning(
                f"failed to deliver search results : {exc!r} (ticket={ticket}, username={username}, query={query})")
        else:
            logger.info(
                f"delivered search results (ticket={ticket}, username={username}, query={query})")
        finally:
            self._search_reply_tasks.remove(task)

    async def _wishlist_job(self, interval: int):
        """Job handling wishlist queries, this method is intended to be run as
        a task. This method will run at the given `interval` (returned by the
        server on start up).
        """
        while True:
            items = self._settings.searches.wishlist

            # Remove all current wishlist searches
            self.requests = {
                ticket: qry for ticket, qry in self.requests.items()
                if qry.search_type != SearchType.WISHLIST
            }

            logger.info(f"starting wishlist search of {len(items)} items")
            # Recreate
            for item in filter(lambda item: item.enabled, items):
                ticket = next(self._ticket_generator)
                self.requests[ticket] = SearchRequest(
                    ticket,
                    item.query,
                    search_type=SearchType.WISHLIST
                )
                self._network.queue_server_messages(
                    WishlistSearch.Request(ticket, item.query)
                )

            await asyncio.sleep(interval)

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    @on_message(SearchInactivityTimeout.Response)
    async def _on_search_inactivity_timeout(self, message: SearchInactivityTimeout.Response, connection):
        self.search_inactivity_timeout = message.timeout

    @on_message(DistributedSearchRequest.Request)
    async def _on_distributed_search_request(
            self, message: DistributedSearchRequest.Request, connection: PeerConnection):

        await self._query_shares_and_reply(message.ticket, message.username, message.query)

    @on_message(DistributedServerSearchRequest.Request)
    async def _on_distributed_server_search_request(
            self, message: DistributedServerSearchRequest.Request, connection: PeerConnection):

        if message.distributed_code != DistributedSearchRequest.Request.MESSAGE_ID:
            logger.warning(f"no handling for server search request with code {message.distributed_code}")
            return

        await self._query_shares_and_reply(message.ticket, message.username, message.query)

    @on_message(ServerSearchRequest.Response)
    async def _on_server_search_request(self, message: ServerSearchRequest.Response, connection: ServerConnection):
        username = self._session.user.name
        if message.username == username:
            return

        await self._query_shares_and_reply(
            message.ticket, message.username, message.query)

    @on_message(FileSearch.Response)
    async def _on_file_search(self, message: FileSearch.Response, connection: ServerConnection):
        """Received when user performs a user or room search"""
        if message.username == self._session.user.name:
            return

        await self._query_shares_and_reply(
            message.ticket, message.username, message.query)

    @on_message(PeerSearchReply.Request)
    async def _on_peer_search_reply(self, message: PeerSearchReply.Request, connection: PeerConnection):
        search_result = SearchResult(
            ticket=message.ticket,
            username=message.username,
            has_free_slots=message.has_slots_free,
            avg_speed=message.avg_speed,
            queue_size=message.queue_size,
            shared_items=message.results,
            locked_results=message.locked_results or []
        )
        try:
            query = self.requests[message.ticket]
        except KeyError:
            logger.warning(f"search reply ticket does not match any search query : {message.ticket}")
        else:
            query.results.append(search_result)
            await self._event_bus.emit(SearchResultEvent(query, search_result))

        await connection.disconnect(reason=CloseReason.REQUESTED)

    @on_message(WishlistInterval.Response)
    async def _on_wish_list_interval(self, message: WishlistInterval.Response, connection):
        self.wishlist_interval = message.interval
        self._cancel_wishlist_task()

        self._wishlist_task = asyncio.create_task(
            self._wishlist_job(message.interval),
            name=f'wishlist-job-{task_counter()}'
        )

    async def _on_state_changed(self, event: ConnectionStateChangedEvent):
        if not isinstance(event.connection, ServerConnection):
            return

        if event.state == ConnectionState.CLOSING:
            self._cancel_wishlist_task()

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        self._session = event.session

    async def _on_session_destroyed(self, event: SessionDestroyedEvent):
        self._session = None

    def _cancel_wishlist_task(self) -> Optional[asyncio.Task]:
        task = self._wishlist_task
        if self._wishlist_task is not None:
            self._wishlist_task.cancel()
            self._wishlist_task = None
            return task
        return None

    async def stop(self) -> List[asyncio.Task]:
        """Cancels all pending tasks

        :return: a list of tasks that have been cancelled so that they can be
            awaited
        """
        cancelled_tasks = []

        for task in self._search_reply_tasks:
            task.cancel()
            cancelled_tasks.append(task)

        if (wishlist_task := self._cancel_wishlist_task()) is not None:
            cancelled_tasks.append(wishlist_task)

        return cancelled_tasks



from dataclasses import dataclass, field
from datetime import datetime
from .user.model import User


@dataclass
class Session:
    user: User
    ip_address: str
    greeting: str
    client_version: int
    minor_version: int
    logged_in_at: datetime = field(default_factory=datetime.utcnow)
    privileges_time_left: int = 0

    @property
    def privileged(self) -> bool:
        return self.privileges_time_left > 0



from __future__ import annotations
from abc import ABC, abstractmethod
import asyncio
import time


INTERVAL = 0.01


class RateLimiter(ABC):

    def __init__(self, limit_bps: int):
        self.limit_bps: int = limit_bps
        self.bucket: int = 0
        self.last_refill: float = 0.0

    @classmethod
    def create_limiter(cls, limit_kbps: int) -> RateLimiter:
        """Creates a new `RateLimiter` instance based on the provided
        `limit_kbps`
        """
        if limit_kbps == 0:
            return UnlimitedRateLimiter()
        else:
            return LimitedRateLimiter(limit_kbps)

    @abstractmethod
    def is_empty(self) -> bool:
        """Returns if the bucket is empty"""
        ...

    @abstractmethod
    def refill(self) -> bool:
        """Refill the bucket"""
        ...

    @abstractmethod
    async def take_tokens(self) -> int:
        """Takes tokens from the bucket"""
        ...

    @abstractmethod
    def add_tokens(self, token_amount: int):
        """Adds tokens to the bucket"""
        ...

    @abstractmethod
    def copy_tokens(self, other: 'RateLimiter'):
        """Copies to tokens from another rate limiter instance"""
        ...


class UnlimitedRateLimiter(RateLimiter):
    MIN_BUCKET_SIZE = 8192

    def __init__(self):
        super().__init__(limit_bps=0)

    def is_empty(self) -> bool:
        return False

    def refill(self) -> bool:
        return False

    async def take_tokens(self) -> int:
        return self.MIN_BUCKET_SIZE

    def add_tokens(self, token_amount: int):
        pass

    def copy_tokens(self, other: RateLimiter):
        pass


class LimitedRateLimiter(RateLimiter):
    MIN_BUCKET_SIZE = 128

    def __init__(self, limit_kbps: int):
        super().__init__(limit_bps=limit_kbps * 1024)

    def is_empty(self) -> bool:
        return self.bucket < self.MIN_BUCKET_SIZE

    def refill(self) -> bool:
        if self.limit_bps == self.bucket:
            return False

        current_time = time.monotonic()
        if self.bucket < self.limit_bps:
            time_passed = current_time - self.last_refill
            new_tokens = (self.limit_bps - self.bucket) * time_passed
            self.add_tokens(int(new_tokens))

        self.last_refill = current_time

        return self.is_empty()

    async def take_tokens(self) -> int:
        while True:
            is_empty = self.refill()
            if not is_empty:
                self.bucket -= self.MIN_BUCKET_SIZE
                return self.MIN_BUCKET_SIZE

            await asyncio.sleep(INTERVAL)

    def add_tokens(self, token_amount: int):
        self.bucket += token_amount
        if self.bucket > self.limit_bps:
            self.bucket = self.limit_bps

    def copy_tokens(self, other: RateLimiter):
        self.add_tokens(other.bucket)
        self.last_refill = other.last_refill



import re

DEFAULT_LISTENING_HOST: str = '0.0.0.0'
PEER_CONNECT_TIMEOUT: float = 10
"""Direct connection timeout"""
PEER_INDIRECT_CONNECT_TIMEOUT: float = 60
"""Indirect connection timeout"""
PEER_INIT_TIMEOUT: float = 5
"""Timeout waiting for Peer Initialization message"""
PEER_READ_TIMEOUT: float = 60
"""Timeout waiting for message on a peer or distributed connection"""
TRANSFER_TIMEOUT: float = 180
"""Timeout waiting for new data on a file connection"""
TRANSFER_REPLY_TIMEOUT = 30
SERVER_CONNECT_TIMEOUT: float = 30
SERVER_LOGIN_TIMEOUT: float = 30
SERVER_PING_INTERVAL: float = 5 * 60
SERVER_RESPONSE_TIMEOUT: float = 30
DISCONNECT_TIMEOUT: float = 10
PATH_SEPERATOR_PATTERN = re.compile(r"[\\/]+")
"""Pattern for splitting/normalizing remote paths"""
UPNP_SEARCH_TIMEOUT: int = 10
POTENTIAL_PARENTS_CACHE_SIZE: int = 20
"""Maximum amount of potential parents stored"""
DEFAULT_COMMAND_TIMEOUT: float = 10



from __future__ import annotations
import logging
import os
import re
from typing import List, TYPE_CHECKING

from ..constants import PATH_SEPERATOR_PATTERN
from ..protocol.primitives import Attribute, FileData

if TYPE_CHECKING:
    from .model import SharedItem


logger = logging.getLogger(__name__)


def normalize_remote_path(path: str) -> str:
    return re.sub(PATH_SEPERATOR_PATTERN, '\\\\', path).rstrip('\\/')


def create_term_pattern(term: str, wildcard=False):
    if wildcard:
        return re.compile(
            r"(?:(?<=\W|_)|^)[^\W_]*{}(?=[\W_]|$)".format(re.escape(term)),
            flags=re.IGNORECASE
        )
    else:
        return re.compile(
            r"(?:(?<=\W|_)|^){}(?=[\W_]|$)".format(re.escape(term)),
            flags=re.IGNORECASE
        )


def convert_item_to_file_data(
        shared_item: SharedItem, use_full_path: bool = True) -> FileData:
    """Convert a `SharedItem` object to a `FileData` object

    :param use_full_path: use the full path of the file as `filename` if
        `True` otherwise use just the filename. Should be `False` when
        generating a shares reply, `True` when generating search reply
    :return: the converted data
    :raise OSError: raised when an error occurred accessing the file
    """
    file_path = shared_item.get_absolute_path()
    file_size = os.path.getsize(file_path)
    file_ext = os.path.splitext(shared_item.filename)[-1][1:]
    if shared_item.attributes:
        attributes = [Attribute(*attr) for attr in shared_item.attributes]
    else:
        attributes = []

    return FileData(
        unknown=1,
        filename=shared_item.get_remote_path() if use_full_path else shared_item.filename,
        filesize=file_size,
        extension=file_ext,
        attributes=attributes
    )


def convert_items_to_file_data(shared_items: List[SharedItem], use_full_path=True) -> List[FileData]:
    """Converts a list of L{SharedItem} instances to a list of L{FileData}
    instances. If an exception occurs when converting the item an error will
    be logged and the item will be omitted from the list
    """
    file_datas = []
    for shared_item in shared_items:
        try:
            file_datas.append(
                convert_item_to_file_data(shared_item, use_full_path=use_full_path)
            )
        except OSError:
            logger.exception(f"failed to convert to result : {shared_item!r}")

    return file_datas


from __future__ import annotations
import asyncio
import logging
from typing import List, Optional

from .base_manager import BaseManager
from .commands import BaseCommand, RC, RT
from .constants import DEFAULT_COMMAND_TIMEOUT
from .distributed import DistributedNetwork
from .events import (
    build_message_map,
    ConnectionStateChangedEvent,
    EventBus,
    MessageReceivedEvent,
    ServerReconnectedEvent,
    SessionDestroyedEvent,
    SessionInitializedEvent,
)
from .exceptions import AioSlskException, AuthenticationError, InvalidSessionError
from .interest.manager import InterestManager
from .shares.cache import SharesCache, SharesNullCache
from .shares.manager import SharesManager
from .network.connection import ConnectionState, ServerConnection
from .network.network import Network
from .peer import PeerManager
from .protocol.messages import Login
from .protocol.primitives import calc_md5
from .room.manager import RoomManager
from .search.manager import SearchManager
from .server import ServerManager
from .session import Session
from .settings import Settings
from .transfer.cache import TransferCache, TransferNullCache
from .transfer.manager import TransferManager
from .user.manager import UserManager
from .utils import ticket_generator


CLIENT_VERSION = 157
MINOR_VERSION = 100


logger = logging.getLogger(__name__)


class SoulSeekClient:
    """SoulSeek client class"""

    def __init__(
            self, settings: Settings,
            shares_cache: Optional[SharesCache] = None, transfer_cache: Optional[TransferCache] = None,
            event_bus: Optional[EventBus] = None):
        self.settings: Settings = settings

        self._stop_event: Optional[asyncio.Event] = None

        self.ticket_generator = ticket_generator()
        self.events: EventBus = event_bus or EventBus()
        self.session: Optional[Session] = None

        self.network: Network = self.create_network()
        self.distributed_network: DistributedNetwork = self.create_distributed_network()

        self.users: UserManager = self.create_user_manager()
        self.rooms: RoomManager = self.create_room_manager()
        self.interests: InterestManager = self.create_interest_manager()

        self.shares: SharesManager = self.create_shares_manager(
            shares_cache or SharesNullCache()
        )
        self.transfers: TransferManager = self.create_transfer_manager(
            transfer_cache or TransferNullCache()
        )
        self.peers: PeerManager = self.create_peer_manager()
        self.searches: SearchManager = self.create_search_manager()
        self.server_manager: ServerManager = self.create_server_manager()

        self.services: List[BaseManager] = [
            self.users,
            self.rooms,
            self.interests,
            self.shares,
            self.transfers,
            self.peers,
            self.searches,
            self.server_manager
        ]

        self._MESSAGE_MAP = build_message_map(self)
        self.register_listeners()

    def get_event_loop(self):
        return asyncio.get_running_loop()

    def register_listeners(self):
        self.events.register(
            MessageReceivedEvent, self._on_message_received)
        self.events.register(
            ConnectionStateChangedEvent, self._on_connection_state_changed)
        self.events.register(
            ServerReconnectedEvent, self._on_server_reconnected)

    async def start(self):
        """Performs a start up of the client consisting of:

        * Calling `load_data` on all defined services
        * Optionally starts a scan of the defined shares
        * Connecting to the server and opening listening ports
        """
        self.get_event_loop().set_exception_handler(self._exception_handler)

        # Allows creating client before actually calling asyncio.run(client.start())
        # see https://stackoverflow.com/questions/55918048/asyncio-semaphore-runtimeerror-task-got-future-attached-to-a-different-loop
        self._stop_event = asyncio.Event()

        await asyncio.gather(*[svc.load_data() for svc in self.services])
        await asyncio.gather(*[svc.start() for svc in self.services])

        if self.settings.shares.scan_on_start:
            asyncio.create_task(self.shares.scan())

        await self.connect()

    async def run_until_stopped(self):
        await self._stop_event.wait()

    async def stop(self):
        """Stops the client, this method consists of:

        * Disconnecting the network and waiting for all connections to close
        * Cancel all pending tasks and waiting for them to complete
        * Write the transfer and shares caches
        """
        logger.info("signaling client to exit")
        self._stop_event.set()

        await self.network.disconnect()

        cancelled_tasks = []
        for service in self.services:
            cancelled_tasks.extend(await service.stop())

        await asyncio.gather(*cancelled_tasks, return_exceptions=True)
        await asyncio.gather(*[svc.store_data() for svc in self.services])

    async def connect(self):
        """Initializes the network by connecting to the server and opening the
        configured listening ports
        """
        await self.network.initialize()

    async def login(
            self, client_version: int = CLIENT_VERSION, minor_version: int = MINOR_VERSION):
        """Performs a logon to the server with the `credentials` defined in the
        `settings`

        :raise AuthenticationError: When authentication failed
        """
        username = self.settings.credentials.username
        password = self.settings.credentials.password

        # Send the login request and wait for the
        await self.network.send_server_messages(
            Login.Request(
                username=username,
                password=password,
                client_version=client_version,
                md5hash=calc_md5(username + password),
                minor_version=minor_version
            )
        )

        response = await self.network.server_connection.receive_message_object()
        if not isinstance(response, Login.Response):
            raise AioSlskException(
                f"expected login response from server but got : {response}")

        if not response.success:
            raise AuthenticationError(
                response.reason,
                f"authentication failed for user : {username}"
            )

        # Notify all listeners that the session has been initialized
        self.session = Session(
            user=self.users.get_user_object(username),
            ip_address=response.ip,
            greeting=response.greeting,
            client_version=client_version,
            minor_version=minor_version
        )
        await self.events.emit(
            SessionInitializedEvent(
                session=self.session,
                raw_message=response
            )
        )

        # Finally start up the reader loop for the server
        self.network.server_connection.start_reader_task()

    def _exception_handler(self, loop, context):
        message = f"unhandled exception on loop {loop!r} : context : {context!r}"
        logger.exception(message, exc_info=context.get('exception', None))

    async def __aenter__(self) -> SoulSeekClient:
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.stop()

    async def __call__(
            self, command: BaseCommand[RC, RT], response: bool = False,
            timeout: float = DEFAULT_COMMAND_TIMEOUT) -> Optional[RT]:
        return await self.execute(command, response=response, timeout=timeout)

    async def execute(
            self, command: BaseCommand[RC, RT], response: bool = False,
            timeout: float = DEFAULT_COMMAND_TIMEOUT) -> Optional[RT]:
        """Execute a `BaseCommand`, see the `commands.py` module for a list of
        possible commands.

        Waiting for a response is optional; the protocol does not always send
        error messages in case of failure. In these cases this method will
        timeout if the command is configured to wait for a response.

        Example waiting for response:

        ..code-block:: python
            from aioslsk.commands import GetUserStatusCommand

            status = await client.execute(
                GetUserStatusCommand('someuser'), reponse=True)

        Example without response:

        ..code-block:: python
            from aioslsk.commands import JoinRoomCommand

            await client.execute(JoinRoomCommand('cool room'))

        :param command: Command class to execute
        :param response: Whether to wait for the response or simply send
            (default: `False`)
        :param timeout: Timeout waiting for response (default: `10`)
        :raise InvalidSessionError: When no logon has been performed
        :return: Optional response depending on how the command was configured
        """
        if self.session is None:
            raise InvalidSessionError("client is not logged in")

        response_future = command.build_expected_response(self)
        if response and response_future:
            response_future = command.build_expected_response(self)
            self.network.register_response_future(response_future)

        try:
            await command.send(self)
        except Exception:
            if response and response_future:
                response_future.cancel()
            raise

        if response and response_future:
            _, response_obj = await asyncio.wait_for(
                response_future, timeout=timeout)
            return command.handle_response(self, response_obj)
        return None

    # Creation methods

    def create_network(self) -> Network:
        return Network(self.settings, self.events)

    def create_user_manager(self) -> UserManager:
        return UserManager(
            self.settings,
            self.events,
            self.network
        )

    def create_room_manager(self) -> RoomManager:
        return RoomManager(
            self.settings,
            self.events,
            self.users,
            self.network
        )

    def create_interest_manager(self) -> InterestManager:
        return InterestManager(
            self.settings,
            self.events,
            self.users,
            self.network
        )

    def create_shares_manager(self, cache: SharesCache) -> SharesManager:
        return SharesManager(
            self.settings,
            self.events,
            self.network,
            cache=cache
        )

    def create_transfer_manager(self, cache: TransferCache) -> TransferManager:
        return TransferManager(
            self.settings,
            self.events,
            self.users,
            self.shares,
            self.network,
            cache=cache
        )

    def create_search_manager(self) -> SearchManager:
        return SearchManager(
            self.settings,
            self.events,
            self.shares,
            self.transfers,
            self.network
        )

    def create_server_manager(self) -> ServerManager:
        return ServerManager(
            self.settings,
            self.events,
            self.network
        )

    def create_peer_manager(self) -> PeerManager:
        return PeerManager(
            self.settings,
            self.events,
            self.users,
            self.shares,
            self.transfers,
            self.network
        )

    def create_distributed_network(self) -> DistributedNetwork:
        return DistributedNetwork(
            self.settings,
            self.events,
            self.network
        )

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    async def _on_connection_state_changed(self, event: ConnectionStateChangedEvent):
        if isinstance(event.connection, ServerConnection):
            if event.state == ConnectionState.CLOSED:
                if session := self.session:
                    session_event = SessionDestroyedEvent(session)
                    self.session = None
                    await self.events.emit(session_event)

    async def _on_server_reconnected(self, event: ServerReconnectedEvent):
        """Automatically logs the user back on in case the server was """
        if self.settings.network.server.reconnect.auto:
            await self.login()



from __future__ import annotations
from functools import partial
from ipaddress import IPv4Address
from typing import List, TYPE_CHECKING
import logging
from async_upnp_client.aiohttp import AiohttpRequester
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.exceptions import UpnpActionResponseError
from async_upnp_client.profiles.igd import IgdDevice
from async_upnp_client.search import async_search
from async_upnp_client.ssdp import SSDP_IP_V4, SSDP_PORT

from ..constants import UPNP_SEARCH_TIMEOUT

if TYPE_CHECKING:
    from ..settings import Settings

logger = logging.getLogger(__name__)


class UPNP:

    def __init__(self, settings: Settings):
        self._settings: Settings = settings
        self._factory: UpnpFactory = UpnpFactory(AiohttpRequester())

    async def search_igd_devices(self, source_ip: str, timeout: int = UPNP_SEARCH_TIMEOUT) -> List[IgdDevice]:
        devices: List[IgdDevice] = []
        logger.info("starting search for IGD devices")
        await async_search(
            partial(self._search_callback, devices),
            source=(source_ip, 0),
            target=(SSDP_IP_V4, SSDP_PORT),
            timeout=timeout
        )
        logger.debug(f"found {len(devices)} IGD devices")
        return devices

    async def _search_callback(self, devices: List[IgdDevice], headers):
        if headers['ST'] not in IgdDevice.DEVICE_TYPES:
            return

        logger.info(f"found Internet Gateway Device : {headers!r}")
        device = await self._factory.async_create_device(headers['LOCATION'])

        devices.append(IgdDevice(device, None))

    async def get_mapped_ports(self, device: IgdDevice):
        entry_count = await device.async_get_port_mapping_number_of_entries()
        if entry_count:
            logger.debug(f"found {entry_count} mapped ports on device {device.name!r}")
            return await self._get_mapped_ports_known(device, entry_count)
        else:
            return await self._get_mapped_ports_unknown(device)

    async def _get_mapped_ports_known(self, device: IgdDevice, count: int):
        entries = []
        for idx in range(count):
            logger.debug(f"getting port map with index {idx} on device {device.name!r}")
            try:
                entry = await device.async_get_generic_port_mapping_entry(idx)
            except Exception as exc:
                logger.debug(f"failed to get entry {idx} on device : {device.name!r}", exc_info=exc)

            else:
                logger.debug(f"got entry for index {idx} on device {device.name!r} : {entry!r}")
                entries.append(entry)

        return entries

    async def _get_mapped_ports_unknown(self, device: IgdDevice):
        """Gets all mapped port entries for the given device when the length of
        the total amount of ports is not known.
        """
        entries = []
        idx = 0
        while True:
            logger.debug(f"getting port map with index {idx} on device {device.name!r}")
            try:
                entry = await device.async_get_generic_port_mapping_entry(idx)

            except UpnpActionResponseError as exc:
                # SpecifiedArrayIndexInvalid
                if not (exc.status == 500 and exc.error_code == 713):
                    logger.debug(f"failed to get entry {idx} on device : {device.name!r}", exc_info=exc)

                break

            except Exception as exc:

                logger.debug(f"failed to get entry {idx} on device : {device.name!r}", exc_info=exc)
                break

            else:
                if entry is None:
                    break
                logger.debug(f"got entry for index {idx} on device {device.name!r} : {entry!r}")
                entries.append(entry)
                idx += 1

        return entries

    async def remove_port_mapping(self, device: IgdDevice, remote_host: str, port: int, protocol: str = 'TCP'):
        try:
            await device.async_delete_port_mapping(
                remote_host=IPv4Address(remote_host),
                external_port=port,
                protocol=protocol
            )
        except Exception as exc:
            logger.warning(
                f"failed to remove port mapping {remote_host}:{port} from device : {device.name!r}",
                exc_info=exc)

    async def map_port(self, device: IgdDevice, source_ip: str, port: int):
        logger.info(f"mapping port {source_ip}:{port} on device {device.name!r}")
        try:
            await device.async_add_port_mapping(
                protocol='TCP',
                remote_host=None,
                external_port=port,
                internal_client=IPv4Address(source_ip),
                internal_port=port,
                enabled=True,
                description='AioSlsk',
                lease_duration=self._settings.network.upnp.lease_duration
            )
        except UpnpActionResponseError as exc:
            logger.warning(f"failed to map port {port} device : {device.name!r}", exc_info=exc)

    async def unmap_port(self, device: IgdDevice, source_ip: str, port: int):
        await device.async_delete_port_mapping(
            remote_host=source_ip,
            external_port=port,
            protocol='TCP'
        )



import os
import re
from typing import List, Tuple

from .utils import split_remote_path


class NamingStrategy:
    NAME = None

    def should_be_applied(self, local_dir: str, local_filename: str) -> bool:
        return True

    def apply(self, remote_path: str, local_dir: str, local_filename: str) -> Tuple[str, str]:
        """Apply the naming changes

        :param remote_path: the original remote path
        :param local_dir: the local directory the file should be written to
        :param local_filename: the local filename the file should be written to
        :return: the `local_dir` and `local_filename` after applying the desired
            changes
        """
        raise NotImplementedError("'apply' should be overwritten by a subclass")


class DefaultNamingStrategy(NamingStrategy):
    """The default naming strategy just uses the filename"""

    def apply(self, remote_path: str, local_dir: str, local_filename: str) -> Tuple[str, str]:
        return local_dir, split_remote_path(remote_path)[-1]


class KeepDirectoryStrategy(NamingStrategy):
    """Keeps the original directory the remote file was in"""

    def apply(self, remote_path: str, local_dir: str, local_filename: str) -> Tuple[str, str]:
        # -1 filename
        # -2 the containing directory
        remote_path_parts = split_remote_path(remote_path)

        # Only a filename (not sure if this can occur)
        if len(remote_path_parts) == 1:
            return local_dir, local_filename

        # Ignore directories starting with '@@' or Windows drives (C:, D:)
        contained_dir = remote_path_parts[-2]
        if contained_dir.startswith('@@'):
            return local_dir, local_filename

        elif re.match(r'[a-zA-Z]{1}:', contained_dir) is not None:
            return local_dir, local_filename

        return os.path.join(local_dir, contained_dir), local_filename


class DuplicateNamingStrategy(NamingStrategy):

    def should_be_applied(self, local_dir: str, local_filename: str) -> bool:
        return os.path.exists(os.path.join(local_dir, local_filename))


class NumberDuplicateStrategy(DuplicateNamingStrategy):
    """Duplicate name strategy that appends a number to the filename in case it
    already exists
    """
    PATTERN = r' \((\d+)\)'

    def apply(self, remote_path: str, local_dir: str, local_filename: str) -> Tuple[str, str]:
        # Find all files which are already numbered
        filename, extension = os.path.splitext(local_filename)
        pattern = re.escape(filename) + self.PATTERN + re.escape(extension)
        indices = []
        for path_file in os.listdir(local_dir):
            if (match := re.match(pattern, path_file)) is not None:
                indices.append(int(match.group(1)))

        # Find the next free index
        next_index = 1
        if indices:
            # +1 to include the max, +2 so that we get the next one in case all
            # indices are already taken
            possible_indices = set(range(min(indices), max(indices) + 2))
            next_index = min(possible_indices - set(indices))

        new_filename = f"{filename} ({next_index}){extension}"
        return local_dir, new_filename


def chain_strategies(strategies: List[NamingStrategy], remote_path: str, local_dir: str) -> Tuple[str, str]:
    """Chains strategies together to find the target location and filename to
    which the file should be written.

    :param strategies: list of strategies to apply
    :param remote_path: the full remote path
    :param local_dir: initial local path, this should be the initial download
        directory
    :return: tuple with the path and filename
    """
    path = local_dir
    filename = None
    for strategy in strategies:
        if strategy.should_be_applied(path, filename):
            path, filename = strategy.apply(remote_path, path, filename)
    return path, filename



import asyncio
import copy
from dataclasses import dataclass
import logging
from typing import Dict, Optional, Set
from weakref import WeakValueDictionary

from ..base_manager import BaseManager
from ..network.connection import ConnectionState, PeerConnection, ServerConnection
from ..events import (
    build_message_map,
    on_message,
    AdminMessageEvent,
    ConnectionStateChangedEvent,
    EventBus,
    KickedEvent,
    MessageReceivedEvent,
    PrivateMessageEvent,
    PrivilegedUsersEvent,
    PrivilegedUserAddedEvent,
    PrivilegesUpdateEvent,
    UserInfoUpdateEvent,
    UserStatsUpdateEvent,
    UserStatusUpdateEvent,
    UserTrackingEvent,
    UserUntrackingEvent,
    SessionInitializedEvent,
    SessionDestroyedEvent,
)
from ..protocol.messages import (
    AddPrivilegedUser,
    AddUser,
    AdminMessage,
    PrivateChatMessage,
    PrivateChatMessageAck,
    CheckPrivileges,
    GetUserStatus,
    GetUserStats,
    Kicked,
    PeerSearchReply,
    PeerUserInfoReply,
    PrivilegedUsers,
    RemoveUser,
    SetStatus,
)
from .model import (
    ChatMessage,
    UploadPermissions,
    User,
    UserStatus,
    TrackingFlag,
)
from ..network.network import Network
from ..settings import Settings
from ..session import Session


logger = logging.getLogger(__name__)


@dataclass
class TrackedUser:
    user: User
    flags: TrackingFlag = TrackingFlag(0)

    def has_add_user_flag(self) -> bool:
        """Returns whether this user has any tracking flags set related to
        AddUser
        """
        add_user_flags = TrackingFlag.FRIEND | TrackingFlag.REQUESTED | TrackingFlag.TRANSFER
        return self.flags & add_user_flags != TrackingFlag(0)


class UserManager(BaseManager):
    """Class responsible for handling user messages and storing users"""

    def __init__(self, settings: Settings, event_bus: EventBus, network: Network):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network

        self._session: Optional[Session] = None

        self._MESSAGE_MAP = build_message_map(self)

        self._users: WeakValueDictionary[str, User] = WeakValueDictionary()
        self._tracked_users: Dict[str, TrackedUser] = dict()
        self._privileged_users: Set[str] = set()

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            MessageReceivedEvent, self._on_message_received)
        self._event_bus.register(
            ConnectionStateChangedEvent, self._on_state_changed)
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)
        self._event_bus.register(
            SessionDestroyedEvent, self._on_session_destroyed)

    @property
    def users(self) -> Dict[str, User]:
        return dict(self._users)

    @property
    def privileged_users(self) -> Set[str]:
        return self._privileged_users

    def get_self(self) -> User:
        """Returns the user object for the current session"""
        return self.get_user_object(self._session.user.name)

    def get_user_object(self, username: str) -> User:
        """Gets a `User` object for given `username`, if the user is not stored
        it will be created and stored

        :param username: Name of the user
        :return: a `User` object
        """
        if username not in self._users:
            user = User(
                name=username,
                privileged=username in self._privileged_users
            )
            self._users[username] = user
        return self._users[username]

    def is_tracked(self, username: str) -> bool:
        return username in self._tracked_users

    def _get_or_create_user(self, username: str) -> User:
        if username not in self._users:
            self._users[username] = User(username)
        return self._users[username]

    def _get_tracked_user_object(self, user: User) -> TrackedUser:
        if user.name not in self._tracked_users:
            self._tracked_users[user.name] = TrackedUser(user)
        return self._tracked_users[user.name]

    def reset_users(self):
        """Performs a reset on all users"""
        self._users = WeakValueDictionary()
        self._tracked_users = dict()
        self._privileged_users = set()

    def set_tracking_flag(self, user: User, flag: TrackingFlag = TrackingFlag.REQUESTED) -> bool:
        """Set given tracking flag for the user. This method returns `True` if
        the user previously had not tracking flags set.
        """
        tracked_user = self._get_tracked_user_object(user)
        had_tracking_flag = tracked_user.flags != TrackingFlag(0)
        tracked_user.flags |= flag
        return not had_tracking_flag

    def unset_tracking_flag(self, user: User, flag: TrackingFlag) -> bool:
        """Unset given tracking flag. This method return `True` if the user has
        not tracking flags left
        """
        tracked_user = self._tracked_users[user.name]
        tracked_user.flags &= ~flag
        return tracked_user.flags != TrackingFlag(0)

    async def track_user(self, username: str, flag: TrackingFlag = TrackingFlag.REQUESTED):
        """Starts tracking a user. The method sends a tracking request to the
        server if it is the first tracking flag being sent or the tracking flag
        is `TrackingFlag.REQUESTED`

        :param username: user to track
        :param flag: tracking flag to add from the user
        """
        user = self.get_user_object(username)
        is_first_flag = self.set_tracking_flag(user, flag)
        if is_first_flag or flag == TrackingFlag.REQUESTED:
            try:
                await self._network.send_server_messages(AddUser.Request(username))
            except Exception:
                self.unset_tracking_flag(user, flag)
                raise

    async def track_friends(self):
        """Starts tracking the users defined in the friends list"""
        tasks = []
        for friend in self._settings.users.friends:
            tasks.append(self.track_user(friend, TrackingFlag.FRIEND))

        await asyncio.gather(*tasks, return_exceptions=True)

    async def untrack_user(self, username: str, flag: TrackingFlag):
        """Removes the given flag from the user and untracks the user (send
        `RemoveUser` message) in case none of the AddUser tracking flags are
        set or the removed tracking flag is `TrackingFlag.REQUESTED`.

        The user will be removed from the tracked users if there are no flags
        left, if there is still a reference left to the user it will remain
        stored

        :param username: user to untrack
        :param flag: tracking flag to remove from the user
        """
        if username not in self._tracked_users:
            return

        user = self._users[username]
        has_flags_left = self.unset_tracking_flag(user, flag)
        if not has_flags_left or flag == TrackingFlag.REQUESTED:
            try:
                await self._network.send_server_messages(RemoveUser.Request(user.name))
            finally:
                # If there's no more tracking done reset the user status and
                # remove the user from the list
                user.status = UserStatus.UNKNOWN

                if username in self._tracked_users:
                    del self._tracked_users[username]
                await self._event_bus.emit(UserUntrackingEvent(user=user))

    @on_message(AdminMessage.Response)
    async def _on_admin_message(self, message: AdminMessage.Response, connection: ServerConnection):
        await self._event_bus.emit(
            AdminMessageEvent(
                message.message,
                raw_message=message
            )
        )

    @on_message(Kicked.Response)
    async def _on_kicked(self, message: Kicked.Response, connection: ServerConnection):
        await self._event_bus.emit(KickedEvent(raw_message=message))

    @on_message(PrivateChatMessage.Response)
    async def _on_private_message(self, message: PrivateChatMessage.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)
        chat_message = ChatMessage(
            id=message.chat_id,
            timestamp=message.timestamp,
            user=user,
            message=message.message,
            is_admin=bool(message.is_admin)
        )

        await self._network.send_server_messages(
            PrivateChatMessageAck.Request(message.chat_id)
        )
        await self._event_bus.emit(
            PrivateMessageEvent(
                message=chat_message,
                raw_message=message
            )
        )

    # State related messages
    @on_message(CheckPrivileges.Response)
    async def _on_check_privileges(self, message: CheckPrivileges.Response, connection: ServerConnection):
        self.privileges_time_left = message.time_left
        self._session.privileges_time_left = message.time_left

        await self._event_bus.emit(
            PrivilegesUpdateEvent(
                time_left=message.time_left,
                raw_message=message
            )
        )

    @on_message(PrivilegedUsers.Response)
    async def _on_privileged_users(self, message: PrivilegedUsers.Response, connection: ServerConnection):
        for user in self._users.values():
            user.privileged = user.name in message.users

        self._privileged_users = set(message.users)

        await self._event_bus.emit(
            PrivilegedUsersEvent(
                users=list(map(self.get_user_object, message.users)),
                raw_message=message
            )
        )

    @on_message(AddPrivilegedUser.Response)
    async def _on_add_privileged_user(self, message: AddPrivilegedUser.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)
        user.privileged = True

        await self._event_bus.emit(PrivilegedUserAddedEvent(user))

    @on_message(AddUser.Response)
    async def _on_add_user(self, message: AddUser.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)
        tracked_user = self._get_tracked_user_object(user)
        if message.exists:
            user.name = message.username
            user.status = UserStatus(message.status)
            user.update_from_user_stats(message.user_stats)
            user.country = message.country_code

            await self._event_bus.emit(
                UserTrackingEvent(
                    user=user,
                    raw_message=message
                )
            )

        else:
            tracked_user.flags = TrackingFlag(0)
            if user.name in self._tracked_users:
                del self._tracked_users[user.name]
            await self._event_bus.emit(UserUntrackingEvent(user=user))

    @on_message(GetUserStatus.Response)
    async def _on_get_user_status(self, message: GetUserStatus.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)

        before = copy.deepcopy(user)

        user.status = UserStatus(message.status)
        user.privileged = message.privileged

        await self._event_bus.emit(
            UserStatusUpdateEvent(
                before=before,
                current=user,
                raw_message=message
            )
        )

    @on_message(GetUserStats.Response)
    async def _on_get_user_stats(self, message: GetUserStats.Response, connection: ServerConnection):
        user = self.get_user_object(message.username)

        before = copy.deepcopy(user)

        user.update_from_user_stats(message.user_stats)

        await self._event_bus.emit(
            UserStatsUpdateEvent(
                before=before,
                current=user,
                raw_message=message
            )
        )

    @on_message(PeerUserInfoReply.Request)
    async def _on_peer_user_info_reply(self, message: PeerUserInfoReply.Request, connection: PeerConnection):
        if not connection.username:
            logger.warning(
                "got PeerUserInfoReply for a connection that wasn't properly initialized")
            return

        user = self.get_user_object(connection.username)

        before = copy.deepcopy(user)

        user.description = message.description
        user.picture = message.picture
        user.upload_slots = message.upload_slots
        user.queue_length = message.queue_size
        user.has_slots_free = message.has_slots_free
        if message.upload_permissions is None:
            user.upload_permissions = UploadPermissions.UNKNOWN
        else:
            user.upload_permissions = UploadPermissions(message.upload_permissions)

        await self._event_bus.emit(
            UserInfoUpdateEvent(
                before=before,
                current=user,
                raw_message=message
            )
        )

    @on_message(PeerSearchReply.Request)
    async def _on_peer_search_reply(self, message: PeerSearchReply.Request, connection: PeerConnection):
        user = self.get_user_object(message.username)
        user.avg_speed = message.avg_speed
        user.queue_length = message.queue_size
        user.has_slots_free = message.has_slots_free

    # Listeners

    async def _on_message_received(self, event: MessageReceivedEvent):
        message = event.message
        if message.__class__ in self._MESSAGE_MAP:
            await self._MESSAGE_MAP[message.__class__](message, event.connection)

    async def _on_state_changed(self, event: ConnectionStateChangedEvent):
        if not isinstance(event.connection, ServerConnection):
            return

        if event.state == ConnectionState.CLOSED:
            self.reset_users()

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        self._session = event.session
        await self._network.send_server_messages(
            CheckPrivileges.Request(),
            SetStatus.Request(UserStatus.ONLINE.value),
        )

        await self.track_user(self._session.user.name, TrackingFlag.FRIEND)

        # Perform AddUser for all in the friendlist
        await self.track_friends()

    async def _on_session_destroyed(self, event: SessionDestroyedEvent):
        self._session = None



from aiofiles import os as asyncos
import asyncio
from functools import partial
import logging
import mutagen
from mutagen.mp3 import BitrateMode
import os
import re
import sys
import time
from typing import Optional, Dict, List, Set, Tuple, Union
import uuid
from weakref import WeakSet

from ..base_manager import BaseManager
from .cache import SharesNullCache, SharesCache
from ..events import (
    EventBus,
    ScanCompleteEvent,
    SessionInitializedEvent,
    SessionDestroyedEvent,
)
from ..exceptions import (
    FileNotFoundError,
    FileNotSharedError,
    SharedDirectoryError,
)
from .model import DirectoryShareMode, SharedDirectory, SharedItem
from ..naming import (
    chain_strategies,
    DefaultNamingStrategy,
    NumberDuplicateStrategy,
)
from ..network.network import Network
from ..protocol.messages import SharedFoldersFiles
from ..protocol.primitives import DirectoryData
from ..search.model import SearchQuery
from ..session import Session
from ..settings import Settings
from .utils import create_term_pattern, convert_items_to_file_data

logger = logging.getLogger(__name__)


_COMPRESSED_FORMATS = [
    'MP3',
    'MP4',
    'ASF',  # WMA
    'OggVorbis'
]
_LOSSLESS_FORMATS = [
    'FLAC',
    'WAVE'
]
_QUERY_CLEAN_PATTERN = re.compile(r"[\W_]")
"""Pattern to remove all non-word/digit characters from a string"""


def scan_directory(shared_directory: SharedDirectory, children: Optional[List[SharedDirectory]] = None) -> Set[SharedItem]:
    """Scans the directory for items to share

    :param shared_directory: `SharedDirectory` instance
    :param children: list of `SharedDirectory` instances, the items in this list
        will not be returned and should be scanned individually
    :return: set of `SharedItem` objects found during the scan
    """
    children = children or []
    shared_items = set()
    for directory, _, files in os.walk(shared_directory.absolute_path):
        # Check if the currently scanned directory is part of any of the given
        # child directories
        abs_dir = os.path.normpath(os.path.abspath(directory))
        if any(child_dir.is_parent_of(abs_dir) for child_dir in children):
            continue

        subdir = os.path.relpath(directory, shared_directory.absolute_path)

        if subdir == '.':
            subdir = ''

        for filename in files:
            filepath = os.path.join(directory, filename)
            try:
                modified = os.path.getmtime(os.path.join(directory, filename))
            except OSError:
                logger.debug(f"could not get modified time for file {filepath!r}")
            else:
                shared_items.add(
                    SharedItem(shared_directory, subdir, filename, modified)
                )

    return shared_items


def extract_attributes(filepath: str) -> List[Tuple[int, int]]:
    """Attempts to extract attributes from the file at `filepath`. If there was
    an error attempting to extract the attributes this method will log a warning
    and return an empty list of attributes
    """
    attributes = []
    try:
        mutagen_file = mutagen.File(filepath)
        if not mutagen_file:
            return []

        if mutagen_file.__class__.__name__ in _COMPRESSED_FORMATS:
            attributes = [
                (0, int(mutagen_file.info.bitrate / 1000), ),
                (1, int(mutagen_file.info.length), )
            ]

            # Figure out bitrate mode if applicable
            bitrate_mode = getattr(mutagen_file, 'bitrate_mode', BitrateMode.UNKNOWN)
            if bitrate_mode == BitrateMode.CBR:
                attributes.append((2, 0, ))
            elif bitrate_mode in (BitrateMode.VBR, BitrateMode.ABR, ):
                attributes.append((2, 1, ))

        elif mutagen_file.__class__.__name__ in _LOSSLESS_FORMATS:
            attributes = [
                (1, int(mutagen_file.info.length), ),
                (4, mutagen_file.info.sample_rate, ),
                (5, mutagen_file.info.bits_per_sample, ),
            ]

    except mutagen.MutagenError as exc:
        logger.warning(
            f"failed retrieve audio file metadata. path={filepath!r}", exc_info=exc)

    return attributes


class SharesManager(BaseManager):
    _ALIAS_LENGTH = 5

    def __init__(
            self, settings: Settings, event_bus: EventBus,
            network: Network, cache: Optional[SharesCache] = None):
        self._settings: Settings = settings
        self._event_bus: EventBus = event_bus
        self._network: Network = network
        self._term_map: Dict[str, WeakSet[SharedItem]] = {}
        self._shared_directories: List[SharedDirectory] = list()
        self._session: Optional[Session] = None

        self.cache: SharesCache = cache if cache else SharesNullCache()
        self.executor = None

        self.naming_strategies = [
            DefaultNamingStrategy(),
            NumberDuplicateStrategy()
        ]

        self.register_listeners()

    def register_listeners(self):
        self._event_bus.register(
            SessionInitializedEvent, self._on_session_initialized)
        self._event_bus.register(
            SessionDestroyedEvent, self._on_session_destroyed)

    @property
    def shared_directories(self):
        return self._shared_directories

    def generate_alias(self, path: str, offset: int = 0) -> str:
        """Generates a directory alias for the given path, this method will be
        called recursively increasing the offset in case the alias is already
        found in the `shared_directories`.

        The hardware address is mixed in to avoid getting the same alias for the
        same directory on different machines. Admittedly this is a lousy
        security measure but hopefully this will prevent easy leaking of files
        in case such issue would occur. Example: 'abcde' is always generated for
        'C:\\' so the attacker could guess where a file is located.

        :param path: the path for which to generate an alias
        :param offset: offset for the value of the initial 5 bytes (default=0)
        :return: a string of 5 alphabetic characters all lowercased
        """
        path_bytes = path.encode('utf8')
        unique_id = uuid.getnode().to_bytes(6, sys.byteorder)

        alias_bytearr = bytearray(
            [unique_id[-1] | offset for _ in range(self._ALIAS_LENGTH)])

        # Chunk the path into pieces of 5
        for c_idx in range(0, len(path_bytes), self._ALIAS_LENGTH):
            chunk = path_bytes[c_idx:c_idx + self._ALIAS_LENGTH]
            # previous_iter_byte XOR unique_id XOR current_iter_byte
            for b_idx, byte in enumerate(chunk):
                alias_bytearr[b_idx] ^= byte ^ unique_id[b_idx]

        # To lowercase
        alias_bytes = bytes([(byte % 26) + 0x61 for byte in alias_bytearr])
        alias_string = alias_bytes.decode('utf8')

        return alias_string

    async def load_data(self):
        self.read_cache()
        self.load_from_settings()

    async def store_data(self):
        self.write_cache()

    def load_from_settings(self):
        """Loads the directories from the settings"""
        for shared_directory in self._settings.shares.directories:
            self.add_shared_directory(
                shared_directory.path,
                share_mode=shared_directory.share_mode,
                users=shared_directory.users
            )

    def read_cache(self):
        """Read the directories from the cache"""
        logger.info("reading directories from cache")
        directories = self.cache.read()
        logger.info(f"read {len(directories)} directories from cache")
        self._shared_directories = directories

    def write_cache(self):
        """Write current shared directories to the cache"""
        logger.info(f"writing {len(self._shared_directories)} directories to cache")
        self.cache.write(self._shared_directories)
        logger.info(f"successfully wrote {len(self._shared_directories)} directories to cache")

    def get_download_directory(self) -> str:
        """Gets the absolute path the to download directory configured from the
        settings

        :return: Absolute path of the value of the `shares.download` setting
        """
        download_dir = self._settings.shares.download
        return os.path.abspath(download_dir)

    async def create_directory(self, absolute_path: str):
        """Ensures the passed directory exists"""
        if not await asyncos.path.exists(absolute_path):
            logger.info(f"creating directory : {absolute_path}")
            await asyncos.makedirs(absolute_path, exist_ok=True)

    def rebuild_term_map(self):
        logger.info("rebuilding term map")
        self._term_map = {}
        for shared_directory in self.shared_directories:
            self._build_term_map(shared_directory)

    def _build_term_map(self, shared_directory: SharedDirectory):
        """Builds a list of valid terms for the given shared directory"""
        for item in shared_directory.items:
            self._add_item_to_term_map(item)
        logger.debug(f"term map contains {len(self._term_map)} terms")

    async def get_shared_item(self, remote_path: str, username: Optional[str] = None) -> SharedItem:
        """Gets a shared item from the cache based on the given file path. If
        the file does not exist in the `shared_items` or the file is present
        in the cache but does not exist on disk a `FileNotFoundError` is raised

        If a `username` is passed this will also check if the file is locked
        and raise a `FileNotSharedError` if the file is not accessible for that
        user

        :param remote_path: the remote_path
        :param username:
        :raise FileNotFoundError: filename was not found in shared_items or was found
            but did not exist on disk
        :raise FileNotSharedError: file is found, but locked for the given
            `username`
        """
        for shared_directory in self._shared_directories:
            try:
                item = shared_directory.get_item_by_remote_path(remote_path)
            except FileNotFoundError:
                pass
            else:
                if not await asyncos.path.exists(item.get_absolute_path()):
                    raise FileNotFoundError(
                        f"file with remote_path {remote_path} found in cache but not on disk"
                    )

                if username and self.is_item_locked(item, username):
                    raise FileNotSharedError(f"File is not shared to user {username}")
                return item
        else:
            raise FileNotFoundError(f"file name {remote_path} not found in cache")

    def add_shared_directory(
            self, shared_directory: str,
            share_mode: DirectoryShareMode = DirectoryShareMode.EVERYONE, users: Optional[List[str]] = None) -> SharedDirectory:
        """Adds a shared directory. This method will call `generate_alias` and
        add the directory to the directory map.

        :param shared_directory: path of the shared directory
        :param share_mode: the share mode for the directory
        :param users: in case the share mode is `USERS`, a list of users to
            share it with
        :return: a `SharedDirectory` object
        """
        # Calc absolute path, generate an alias and store it
        abs_directory = os.path.normpath(os.path.abspath(shared_directory))
        alias = self.generate_alias(abs_directory)

        # Check if we have an existing shared directory, otherwise return it
        directory_object = SharedDirectory(
            shared_directory,
            abs_directory,
            alias,
            share_mode=share_mode,
            users=users or []
        )
        for shared_dir in self._shared_directories:
            if shared_dir == directory_object:
                return shared_dir

        # If the new directory is a child of an existing directory, move items
        # to the child directory and remove them from the parent directory
        parents = self._get_parent_directories(directory_object)
        if parents:
            parent = parents[-1]
            children = parent.get_items_for_directory(directory_object)
            directory_object.items |= children
            parent.items -= children

        self._shared_directories.append(directory_object)
        return directory_object

    def remove_shared_directory(self, directory: SharedDirectory):
        """Removes the given shared directory. If the directory was a
        subdirectory of another shared directory its items will be moved into
        that directory

        :param shared_directory: `SharedDirectory` instance to remove
        :raise SharedDirectoryError: raised when the passed `shared_directory`
            was not added to the manager
        """
        if directory not in self._shared_directories:
            raise SharedDirectoryError(
                "attempted to remove directory which was not added to the manager"
            )

        self._shared_directories.remove(directory)

        parents = self._get_parent_directories(directory)
        # If the directory has a parent directory, move all items into that
        # directory
        if parents:
            parent = parents[-1]
            parent.items |= directory.items

        self._cleanup_term_map()

    async def scan_directory_files(self, shared_directory: SharedDirectory):
        """Scans the files for the given `shared_directory`

        :param shared_directory: `SharedDirectory` instance to scan
        :raise SharedDirectoryError: raised when the passed `shared_directory`
            was not added to the manager
        """
        loop = asyncio.get_running_loop()

        if shared_directory not in self._shared_directories:
            raise SharedDirectoryError(
                "attempted to scan directory which was not added to the manager"
            )

        logger.info(f"scheduling scan for directory : {shared_directory!r})")
        try:
            shared_items: Set[SharedItem] = await loop.run_in_executor(
                self.executor,
                partial(
                    scan_directory,
                    shared_directory,
                    children=self._get_child_directories(shared_directory)
                )
            )
        except Exception:
            logger.exception(f"exception scanning directory : {shared_directory!r}")
        else:
            logger.debug(f"scan found {len(shared_items)} files for directory {shared_directory!r}")

            # Adds all new items to the directory items
            shared_directory.items |= shared_items
            # Remove all items from the cache that weren't in the returned items
            # set
            shared_directory.items -= (shared_directory.items ^ shared_items)

        self._build_term_map(shared_directory)
        self._cleanup_term_map()

    async def scan_directory_file_attributes(self, shared_directory: SharedDirectory):
        """Scans the file attributes for files in the given `shared_directory`.
        only files that do not yet have attributes will be scanned

        The results of the scan are handled internally and are automatically to
        the `SharedItem` object for which the scan was performed

        :param shared_directory: `SharedDirectory` instance for which the files
            need to be scanned
        :return: List of futures for each file that needs to be scanned
        """
        def extract_item_attributes(item: SharedItem):
            return item, extract_attributes(item.get_absolute_path())

        loop = asyncio.get_running_loop()

        # Schedule the items on the executor
        extract_futures: List[asyncio.Future] = []
        for item in shared_directory.items:
            if item.attributes is None:
                future = loop.run_in_executor(
                    self.executor,
                    partial(extract_item_attributes, item)
                )
                extract_futures.append(future)

        logger.debug(
            f"scheduled {len(extract_futures)} / {len(shared_directory.items)} items "
            f"for attribute extracting for directory {shared_directory}")

        for future in asyncio.as_completed(extract_futures):
            try:
                item, attributes = await future
            except Exception:
                logger.warning(f"exception fetching shared item attributes")
            else:
                item.attributes = attributes

    async def scan(self):
        """Scan the files and their attributes for all directories currently
        defined in the `shared_directories`

        This method will emit a `ScanCompleteEvent` on the event bus and report
        the shares to the server
        """
        start_time = time.perf_counter()

        files_tasks = [
            self.scan_directory_files(shared_directory)
            for shared_directory in self._shared_directories
        ]
        await asyncio.gather(*files_tasks, return_exceptions=True)

        attribute_futures = [
            self.scan_directory_file_attributes(shared_directory)
            for shared_directory in self._shared_directories
        ]
        await asyncio.gather(*attribute_futures, return_exceptions=True)

        logger.info(f"completed scan in {time.perf_counter() - start_time} seconds")
        folder_count, file_count = self.get_stats()
        await self._event_bus.emit(
            ScanCompleteEvent(folder_count, file_count)
        )

        await self.report_shares()

    async def get_filesize(self, shared_item: SharedItem) -> int:
        return await asyncos.path.getsize(shared_item.get_absolute_path())

    def query(self, query: Union[str, SearchQuery], username: Optional[str] = None) -> Tuple[List[SharedItem], List[SharedItem]]:
        """Performs a query on the `shared_directories` returning the matching
        items. If `username` is passed this method will return a list of
        visible results and list of locked results. If `None` the second list
        will always be empty.

        This method makes a first pass using the built in term map and filters
        the remaining results using regular expressions.

        :param query: the query to perform on the shared directories
        :param username: optionally the username of the user making the query.
            This is used to determine locked results, if not given the locked
            results list will be empty
        :return: two lists of visible results and locked results
        """
        if not isinstance(query, SearchQuery):
            search_query = SearchQuery.parse(query)
        else:
            search_query = query

        # Ignore if no valid include or wildcard terms are given
        if not search_query.has_inclusion_terms():
            return [], []

        # First round using the term map
        include_terms = []
        for term in search_query.include_terms:
            subterms = re.split(_QUERY_CLEAN_PATTERN, term)
            for subterm in subterms:
                if not subterm:
                    continue

                if subterm not in self._term_map:  # Optimization
                    return [], []

                include_terms.append(subterm)

        for term in search_query.wildcard_terms:
            subterms = re.split(_QUERY_CLEAN_PATTERN, term)
            for idx, subterm in enumerate(subterms):
                if not subterm:
                    continue

                if idx == 0:
                    matching_terms = [
                        map_term for map_term in self._term_map.keys()
                        if map_term.endswith(subterm)
                    ]

                    if not matching_terms:  # Optimization
                        return [], []

                    include_terms.extend(matching_terms)
                else:
                    if subterm not in self._term_map:  # Optimization
                        return [], []

                    include_terms.append(subterm)

        found_items = set(self._term_map[include_terms[0]])
        for include_term in include_terms:
            found_items &= set(self._term_map[include_term])

        # Regular expressions on the remnants

        for include_term in search_query.include_terms:
            to_remove = set()
            for item in found_items:
                pattern = create_term_pattern(include_term, wildcard=False)
                if not re.search(pattern, item.get_query_path()):
                    to_remove.add(item)
            found_items -= to_remove

        for wildcard_term in search_query.wildcard_terms:
            to_remove = set()
            for item in found_items:
                pattern = create_term_pattern(wildcard_term, wildcard=True)
                if not re.search(pattern, item.get_query_path()):
                    to_remove.add(item)
            found_items -= to_remove

        for exclude_term in search_query.exclude_terms:
            to_remove = set()
            for item in found_items:
                pattern = create_term_pattern(exclude_term, wildcard=False)
                if re.search(pattern, item.get_query_path()):
                    to_remove.add(item)
            found_items -= to_remove

        # Order by visible and locked results when username is given
        if username:
            visible_results = []
            locked_results = []
            for item in found_items:
                if self.is_item_locked(item, username):
                    locked_results.append(item)
                else:
                    visible_results.append(item)
            return visible_results, locked_results
        else:
            return list(found_items), []

    def get_stats(self) -> Tuple[int, int]:
        """Gets the total amount of shared directories and files.

        :return: directory and file count as a `tuple`
        """
        file_count = sum(
            len(directory.items) for directory in self._shared_directories
        )
        dir_count = sum(
            len(set(item.subdir for item in directory.items))
            for directory in self._shared_directories
        )
        return dir_count, file_count

    def calculate_download_path(self, remote_path: str) -> Tuple[str, str]:
        """Calculates the local download path for a remote path returned by
        another peer.

        :return: tuple of the directory and file name
        """
        download_dir = self.get_download_directory()

        return chain_strategies(
            self.naming_strategies,
            remote_path,
            download_dir
        )

    def get_shared_directories_for_user(self, username: str) -> Tuple[List[SharedDirectory], List[SharedDirectory]]:
        public_dirs = []
        locked_dirs = []
        for shared_dir in self._shared_directories:
            if self.is_directory_locked(shared_dir, username):
                locked_dirs.append(shared_dir)
            else:
                public_dirs.append(shared_dir)

        return public_dirs, locked_dirs

    def create_shares_reply(self, username: str) -> Tuple[List[DirectoryData], List[DirectoryData]]:
        """Creates a complete list of the currently shared items as a reply to
        a `PeerSharesRequest` messages

        :param username: username of the user requesting the shares reply, this
            is used to determine the locked results
        :return: tuple with two lists: public directories and locked directories
        """
        def list_unique_directories(directories: List[SharedDirectory]) -> Dict[str, List[SharedItem]]:
            # Sort files under unique directories by path
            response_dirs: Dict[str, List[SharedItem]] = {}
            for directory in directories:
                for item in directory.items:
                    dir_path = item.get_remote_directory_path()
                    if dir_path in response_dirs:
                        response_dirs[dir_path].append(item)
                    else:
                        response_dirs[dir_path] = [item, ]
            return response_dirs

        def convert_to_directory_shares(directory_map: Dict[str, List[SharedItem]]) -> List[DirectoryData]:
            public_shares = []
            for directory, files in directory_map.items():
                public_shares.append(
                    DirectoryData(
                        name=directory,
                        files=convert_items_to_file_data(files, use_full_path=False)
                    )
                )
            return public_shares

        visible_dirs, locked_dirs = self.get_shared_directories_for_user(username)
        visible_shares = convert_to_directory_shares(
            list_unique_directories(visible_dirs))
        locked_shares = convert_to_directory_shares(
            list_unique_directories(locked_dirs))

        return visible_shares, locked_shares

    def create_directory_reply(self, remote_directory: str) -> List[DirectoryData]:
        """Lists directory data as a response to a directory request.

        :param remote_directory: remote path of the directory
        """
        remote_directory = remote_directory.rstrip('\\/')
        response_dirs: Dict[str, List[SharedItem]] = {}
        for shared_dir in self._shared_directories:
            for item in shared_dir.items:
                item_dir = item.get_remote_directory_path()
                if item_dir != remote_directory:
                    continue

                dir_name = shared_dir.get_remote_path()
                if dir_name in response_dirs:
                    response_dirs[dir_name].append(item)
                else:
                    response_dirs[dir_name] = [item, ]

        reply = []
        for directory, files in response_dirs.items():
            reply.append(
                DirectoryData(
                    name=directory,
                    files=convert_items_to_file_data(files, use_full_path=False)
                )
            )

        return reply

    def _add_item_to_term_map(self, item: SharedItem):
        path = (item.subdir + "/" + item.filename).lower()
        terms = re.split(_QUERY_CLEAN_PATTERN, path)
        for term in terms:
            if not term:
                continue

            if term not in self._term_map:
                self._term_map[term] = WeakSet()
            self._term_map[term].add(item)

    def _cleanup_term_map(self):
        self._term_map = {
            term: values for term, values in self._term_map.items()
            if len(values) > 0
        }

    def _get_parent_directories(self, shared_directory: SharedDirectory) -> List[SharedDirectory]:
        """Returns a list of parent shared directories. The parent directories
        will be sorted by length of the absolute path (longest last)
        """
        parent_dirs = [
            directory for directory in self._shared_directories
            if directory != shared_directory and directory.is_parent_of(shared_directory)
        ]
        return sorted(parent_dirs, key=lambda d: len(d.absolute_path))

    def _get_child_directories(self, shared_directory: SharedDirectory):
        """Returns a list of child directories"""
        children = [
            directory for directory in self._shared_directories
            if directory != shared_directory and directory.is_child_of(shared_directory)
        ]
        return children

    def is_directory_locked(self, directory: SharedDirectory, username: str) -> bool:
        """Checks if the shared directory is locked for the given `username`"""
        if directory.share_mode == DirectoryShareMode.FRIENDS:
            return username not in self._settings.users.friends
        elif directory.share_mode == DirectoryShareMode.USERS:
            return username not in directory.users
        return False

    def is_item_locked(self, item: SharedItem, username: str) -> bool:
        """Checks if the shared item is locked for the given `username`"""
        return self.is_directory_locked(item.shared_directory, username)

    async def report_shares(self):
        """Reports the shares amount to the server"""
        if not self._session:
            logger.warning("attempted to report shares without valid session")
            return

        folder_count, file_count = self.get_stats()
        logger.debug(f"reporting shares ({folder_count=}, {file_count=})")
        await self._network.send_server_messages(
            SharedFoldersFiles.Request(
                shared_folder_count=folder_count,
                shared_file_count=file_count
            )
        )

    async def _on_session_initialized(self, event: SessionInitializedEvent):
        self._session = event.session
        await self.report_shares()

    async def _on_session_destroyed(self, event: SessionDestroyedEvent):
        self._session = None
